from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional, Sequence, Tuple

from dbt_semantic_interfaces.call_parameter_sets import TimeDimensionCallParameterSet
from dbt_semantic_interfaces.naming.keywords import METRIC_TIME_ELEMENT_NAME
from dbt_semantic_interfaces.references import SemanticModelReference, TimeDimensionReference
from dbt_semantic_interfaces.type_enums import TimeGranularity
from typing_extensions import override

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.model.semantics.linkable_element_set import LinkableElementSet
from metricflow_semantics.naming.object_builder_scheme import ObjectBuilderNamingScheme
from metricflow_semantics.query.group_by_item.candidate_push_down.push_down_visitor import (
    PushDownResult,
    _PushDownGroupByItemCandidatesVisitor,
)
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import WhereFilterLocation
from metricflow_semantics.query.group_by_item.resolution_dag.dag import GroupByItemResolutionDag, ResolutionDagSinkNode
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.group_by_item_resolver.ambiguous_group_by_item import AmbiguousGroupByItemIssue
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryResolutionIssueSet,
)
from metricflow_semantics.query.suggestion_generator import QueryItemSuggestionGenerator
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.patterns.metric_time_default_granularity import MetricTimeDefaultGranularityPattern
from metricflow_semantics.specs.patterns.minimum_time_grain import MinimumTimeGrainPattern
from metricflow_semantics.specs.patterns.no_group_by_metric import NoGroupByMetricPattern
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.patterns.typed_patterns import TimeDimensionPattern
from metricflow_semantics.specs.spec_set import InstanceSpecSet, group_specs_by_type

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class GroupByItemResolution(SemanticModelDerivation):
    """Result object that contains matching spec for a potentially ambiguous input to a query.

    e.g. "TimeDimension('metric_time')" -> TimeDimensionSpec('metric_time', DAY)

    Issues found while resolving the input pattern are returned in issue_set.
    """

    # If the spec is None, then the pattern couldn't be resolved
    spec: Optional[LinkableInstanceSpec]
    linkable_element_set: LinkableElementSet
    issue_set: MetricFlowQueryResolutionIssueSet

    @property
    @override
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        return self.linkable_element_set.derived_from_semantic_models


@dataclass(frozen=True)
class AvailableGroupByItemsResolution:
    """Result object that contains the available group-by-items for a query."""

    specs: Tuple[LinkableInstanceSpec, ...]
    issue_set: MetricFlowQueryResolutionIssueSet


class GroupByItemResolver:
    """Resolves group-by items for potentially ambiguous inputs that are specified in queries / filters."""

    def __init__(  # noqa: D107
        self,
        manifest_lookup: SemanticManifestLookup,
        resolution_dag: GroupByItemResolutionDag,
    ) -> None:
        self._manifest_lookup = manifest_lookup
        self._resolution_dag = resolution_dag

    def resolve_matching_item_for_querying(
        self,
        spec_pattern: SpecPattern,
        suggestion_generator: Optional[QueryItemSuggestionGenerator],
    ) -> GroupByItemResolution:
        """Returns the spec that corresponds to the one described by spec_pattern and is valid for the query.

        For queries, if the pattern matches to a spec for the same element at different grains, the spec with the finest
        common grain is returned, unless the spec is metric_time, in which case the default grain is returned.
        """
        push_down_visitor = _PushDownGroupByItemCandidatesVisitor(
            manifest_lookup=self._manifest_lookup,
            source_spec_patterns=(spec_pattern, NoGroupByMetricPattern()),
            group_by_item_resolver=self,
            suggestion_generator=suggestion_generator,
        )

        push_down_result: PushDownResult = self._resolution_dag.sink_node.accept(push_down_visitor)

        if push_down_result.candidate_set.num_candidates == 0:
            return GroupByItemResolution(
                spec=None,
                linkable_element_set=LinkableElementSet(),
                issue_set=push_down_result.issue_set,
            )

        for candidate_filter in (
            # Default pattern must come first to avoid removing default grain options prematurely.
            MetricTimeDefaultGranularityPattern(push_down_result.max_metric_default_time_granularity),
            MinimumTimeGrainPattern(),
        ):
            push_down_result = push_down_result.filter_candidates_by_pattern(candidate_filter)

        logger.debug(
            LazyFormat(
                lambda: f"Spec pattern:\n"
                f"{indent(mf_pformat(spec_pattern))}\n"
                f"was resolved to:\n"
                f"{indent(mf_pformat(push_down_result.candidate_set.specs))}"
            )
        )
        if push_down_result.candidate_set.num_candidates > 1:
            return GroupByItemResolution(
                spec=None,
                linkable_element_set=LinkableElementSet(),
                issue_set=push_down_result.issue_set.add_issue(
                    AmbiguousGroupByItemIssue.from_parameters(
                        candidate_set=push_down_result.candidate_set,
                        query_resolution_path=MetricFlowQueryResolutionPath.from_path_item(
                            self._resolution_dag.sink_node
                        ),
                    )
                ),
            )

        return GroupByItemResolution(
            spec=push_down_result.candidate_set.specs[0],
            linkable_element_set=push_down_result.candidate_set.linkable_element_set,
            issue_set=push_down_result.issue_set,
        )

    def resolve_matching_item_for_filters(
        self,
        input_str: str,
        spec_pattern: SpecPattern,
        resolution_node: ResolutionDagSinkNode,
        filter_location: WhereFilterLocation,
    ) -> GroupByItemResolution:
        """Returns the spec that matches the spec_pattern associated with the filter in the given node.

        The input_str is a string representation of the input for use in generating suggestions for errors.

        e.g. if the query has a where filter like "{{ TimeDimension('metric_time') }} = '2020-01-01'", then to resolve
        the spec for TimeDimension('metric_time'), the resolution node would be a QueryGroupByItemResolutionNode() from
        the DAG used in the initialization of this resolver.
        """
        suggestion_generator = QueryItemSuggestionGenerator(
            input_naming_scheme=ObjectBuilderNamingScheme(),
            input_str=input_str,
            candidate_filters=QueryItemSuggestionGenerator.FILTER_ITEM_CANDIDATE_FILTERS,
        )

        push_down_visitor = _PushDownGroupByItemCandidatesVisitor(
            manifest_lookup=self._manifest_lookup,
            source_spec_patterns=(spec_pattern,),
            group_by_item_resolver=self,
            suggestion_generator=suggestion_generator,
            filter_location=filter_location,
        )

        push_down_result: PushDownResult = resolution_node.accept(push_down_visitor)

        for candidate_filter in (
            # Default pattern must come first to avoid removing default grain options prematurely.
            MetricTimeDefaultGranularityPattern(push_down_result.max_metric_default_time_granularity),
            MinimumTimeGrainPattern(),
        ):
            push_down_result = push_down_result.filter_candidates_by_pattern(candidate_filter)

        if push_down_result.candidate_set.num_candidates == 0:
            return GroupByItemResolution(
                spec=None,
                linkable_element_set=LinkableElementSet(),
                issue_set=push_down_result.issue_set,
            )

        if push_down_result.candidate_set.num_candidates > 1:
            return GroupByItemResolution(
                spec=None,
                linkable_element_set=LinkableElementSet(),
                issue_set=push_down_result.issue_set.add_issue(
                    AmbiguousGroupByItemIssue.from_parameters(
                        candidate_set=push_down_result.candidate_set,
                        query_resolution_path=MetricFlowQueryResolutionPath.from_path_item(
                            self._resolution_dag.sink_node
                        ),
                    )
                ),
            )

        return GroupByItemResolution(
            spec=push_down_result.candidate_set.specs[0],
            linkable_element_set=push_down_result.candidate_set.linkable_element_set,
            issue_set=push_down_result.issue_set,
        )

    def resolve_available_items(
        self,
        resolution_node: Optional[ResolutionDagSinkNode] = None,
        source_spec_patterns: Sequence[SpecPattern] = (),
    ) -> AvailableGroupByItemsResolution:
        """Return all available group-by-items at a given node.

        By default, the query node is used, so this will return the available group-by-items that can be used in a
        query.
        """
        if resolution_node is None:
            resolution_node = self._resolution_dag.sink_node

        push_down_visitor = _PushDownGroupByItemCandidatesVisitor(
            manifest_lookup=self._manifest_lookup,
            source_spec_patterns=source_spec_patterns,
            group_by_item_resolver=self,
            suggestion_generator=None,
        )

        push_down_result: PushDownResult = resolution_node.accept(push_down_visitor)

        return AvailableGroupByItemsResolution(
            specs=tuple(push_down_result.candidate_set.specs),
            issue_set=push_down_result.issue_set,
        )

    def resolve_min_metric_time_grain(self) -> TimeGranularity:
        """Returns the finest base time grain of metric_time for querying."""
        metric_time_grain_resolution = self.resolve_matching_item_for_querying(
            spec_pattern=TimeDimensionPattern.from_call_parameter_set(
                TimeDimensionCallParameterSet(
                    entity_path=(),
                    time_dimension_reference=TimeDimensionReference(element_name=METRIC_TIME_ELEMENT_NAME),
                )
            ),
            suggestion_generator=None,
        )
        metric_time_spec_set = (
            group_specs_by_type((metric_time_grain_resolution.spec,))
            if metric_time_grain_resolution.spec is not None
            else InstanceSpecSet.empty_instance()
        )
        if len(metric_time_spec_set.time_dimension_specs) != 1:
            raise RuntimeError(
                f"The grain for {repr(METRIC_TIME_ELEMENT_NAME)} could not be resolved. Got spec "
                f"{metric_time_grain_resolution.spec} and issues:\n\n"
                f"{indent(mf_pformat(metric_time_grain_resolution.issue_set))}"
            )
        return metric_time_spec_set.time_dimension_specs[0].time_granularity.base_granularity



from __future__ import annotations

import itertools
import logging
from typing import List, Optional, Sequence

import jinja2
from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilterIntersection
from dbt_semantic_interfaces.protocols import WhereFilter, WhereFilterIntersection

from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_location import WhereFilterLocation
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
)
from metricflow_semantics.specs.column_assoc import ColumnAssociationResolver
from metricflow_semantics.specs.linkable_spec_set import LinkableSpecSet
from metricflow_semantics.specs.rendered_spec_tracker import RenderedSpecTracker
from metricflow_semantics.specs.where_filter.where_filter_dimension import WhereFilterDimensionFactory
from metricflow_semantics.specs.where_filter.where_filter_entity import WhereFilterEntityFactory
from metricflow_semantics.specs.where_filter.where_filter_metric import WhereFilterMetricFactory
from metricflow_semantics.specs.where_filter.where_filter_spec import WhereFilterSpec
from metricflow_semantics.specs.where_filter.where_filter_time_dimension import WhereFilterTimeDimensionFactory
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet

logger = logging.getLogger(__name__)


class RenderSqlTemplateException(Exception):  # noqa: D101
    pass


class WhereSpecFactory:
    """Renders the SQL template in the WhereFilter and converts it to a WhereFilterSpec."""

    def __init__(  # noqa: D107
        self,
        column_association_resolver: ColumnAssociationResolver,
        spec_resolution_lookup: FilterSpecResolutionLookUp,
        semantic_model_lookup: SemanticModelLookup,
    ) -> None:
        self._column_association_resolver = column_association_resolver
        self._spec_resolution_lookup = spec_resolution_lookup
        self._semantic_model_lookup = semantic_model_lookup

    def create_from_where_filter(  # noqa: D102
        self,
        filter_location: WhereFilterLocation,
        where_filter: WhereFilter,
    ) -> WhereFilterSpec:
        return self.create_from_where_filter_intersection(
            filter_location=filter_location,
            filter_intersection=PydanticWhereFilterIntersection(where_filters=[where_filter]),
        )[0]

    def create_from_where_filter_intersection(  # noqa: D102
        self,
        filter_location: WhereFilterLocation,
        filter_intersection: Optional[WhereFilterIntersection],
    ) -> Sequence[WhereFilterSpec]:
        if filter_intersection is None:
            return ()

        filter_specs: List[WhereFilterSpec] = []

        for where_filter in filter_intersection.where_filters:
            rendered_spec_tracker = RenderedSpecTracker()
            dimension_factory = WhereFilterDimensionFactory(
                column_association_resolver=self._column_association_resolver,
                spec_resolution_lookup=self._spec_resolution_lookup,
                where_filter_location=filter_location,
                rendered_spec_tracker=rendered_spec_tracker,
            )
            time_dimension_factory = WhereFilterTimeDimensionFactory(
                column_association_resolver=self._column_association_resolver,
                spec_resolution_lookup=self._spec_resolution_lookup,
                where_filter_location=filter_location,
                rendered_spec_tracker=rendered_spec_tracker,
                custom_granularity_names=self._semantic_model_lookup.custom_granularity_names,
            )
            entity_factory = WhereFilterEntityFactory(
                column_association_resolver=self._column_association_resolver,
                spec_resolution_lookup=self._spec_resolution_lookup,
                where_filter_location=filter_location,
                rendered_spec_tracker=rendered_spec_tracker,
            )
            metric_factory = WhereFilterMetricFactory(
                column_association_resolver=self._column_association_resolver,
                spec_resolution_lookup=self._spec_resolution_lookup,
                where_filter_location=filter_location,
                rendered_spec_tracker=rendered_spec_tracker,
            )
            try:
                # If there was an error with the template, it should have been caught while resolving the specs for
                # the filters during query resolution.
                where_sql = jinja2.Template(where_filter.where_sql_template, undefined=jinja2.StrictUndefined).render(
                    {
                        "Dimension": dimension_factory.create,
                        "TimeDimension": time_dimension_factory.create,
                        "Entity": entity_factory.create,
                        "Metric": metric_factory.create,
                    }
                )
            except (jinja2.exceptions.UndefinedError, jinja2.exceptions.TemplateSyntaxError) as e:
                raise RenderSqlTemplateException(
                    f"Error while rendering Jinja template:\n{where_filter.where_sql_template}"
                ) from e
            rendered_specs = tuple(result[0] for result in rendered_spec_tracker.rendered_specs_to_elements)
            linkable_elements = tuple(
                itertools.chain.from_iterable(result[1] for result in rendered_spec_tracker.rendered_specs_to_elements)
            )
            filter_specs.append(
                WhereFilterSpec(
                    where_sql=where_sql,
                    bind_parameters=SqlBindParameterSet(),
                    linkable_spec_set=LinkableSpecSet.create_from_specs(rendered_specs),
                    linkable_element_unions=tuple(linkable_element.as_union for linkable_element in linkable_elements),
                )
            )

        return filter_specs



from __future__ import annotations

import logging
import time
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, FrozenSet, List, Optional, Sequence, Set, Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols.dimension import Dimension, DimensionType
from dbt_semantic_interfaces.protocols.metric import Metric
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    EntityReference,
    MeasureReference,
    MetricReference,
    SemanticModelReference,
    TimeDimensionReference,
)
from dbt_semantic_interfaces.type_enums import MetricType
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from dbt_semantic_interfaces.validations.unique_valid_name import MetricFlowReservedKeywords

from metricflow_semantics.errors.error_classes import UnknownMetricLinkingError
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.model.linkable_element_property import LinkableElementProperty
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.model.semantics.linkable_element import (
    ElementPathKey,
    LinkableDimension,
    LinkableEntity,
    LinkableMetric,
    MetricSubqueryJoinPathElement,
    SemanticModelJoinPath,
    SemanticModelJoinPathElement,
    SemanticModelToMetricSubqueryJoinPath,
)
from metricflow_semantics.model.semantics.linkable_element_set import LinkableElementSet
from metricflow_semantics.model.semantics.semantic_model_helper import SemanticModelHelper
from metricflow_semantics.model.semantics.semantic_model_join_evaluator import SemanticModelJoinEvaluator
from metricflow_semantics.specs.time_dimension_spec import DEFAULT_TIME_GRANULARITY
from metricflow_semantics.time.granularity import ExpandedTimeGranularity
from metricflow_semantics.time.time_spine_source import TimeSpineSource

if TYPE_CHECKING:
    from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup


logger = logging.getLogger(__name__)


class ValidLinkableSpecResolver:
    """Figures out what linkable specs are valid for a given metric.

    e.g. Can you query the metric "bookings" by "listing__country_latest"?
    """

    def __init__(
        self,
        semantic_manifest: SemanticManifest,
        semantic_model_lookup: SemanticModelLookup,
        max_entity_links: int,
    ) -> None:
        """Constructor.

        Args:
            semantic_manifest: the model to use.
            semantic_model_lookup: used to look up entities for a semantic model.
            max_entity_links: the maximum number of joins to do when computing valid elements.
        """
        self._semantic_manifest = semantic_manifest
        self._semantic_model_lookup = semantic_model_lookup
        # Sort semantic models by name for consistency in building derived objects.
        self._semantic_models = sorted(self._semantic_manifest.semantic_models, key=lambda x: x.name)
        self._join_evaluator = SemanticModelJoinEvaluator(semantic_model_lookup)
        self._time_spine_sources = TimeSpineSource.build_standard_time_spine_sources(self._semantic_manifest)
        self._custom_granularities = TimeSpineSource.build_custom_granularities(list(self._time_spine_sources.values()))

        assert max_entity_links >= 0
        self._max_entity_links = max_entity_links

        # Map measures / entities to semantic models that contain them.
        self._entity_to_semantic_model: Dict[str, List[SemanticModel]] = defaultdict(list)
        self._semantic_model_reference_to_semantic_model: Dict[SemanticModelReference, SemanticModel] = {}
        for semantic_model in self._semantic_models:
            for entity in semantic_model.entities:
                self._entity_to_semantic_model[entity.reference.element_name].append(semantic_model)
            self._semantic_model_reference_to_semantic_model[semantic_model.reference] = semantic_model

        self._metric_to_linkable_element_sets: Dict[str, List[LinkableElementSet]] = {}
        self._metric_references_to_metrics: Dict[MetricReference, Metric] = {}
        self._joinable_metrics_for_entities: Dict[EntityReference, Set[MetricSubqueryJoinPathElement]] = defaultdict(
            set
        )

        # Cache for `_get_joined_elements()`.
        self._semantic_model_reference_to_joined_elements: Dict[SemanticModelReference, LinkableElementSet] = {}

        start_time = time.time()
        for metric in self._semantic_manifest.metrics:
            self._metric_references_to_metrics[MetricReference(metric.name)] = metric
            linkable_sets_for_measure = []
            for measure in metric.measure_references:
                if metric.type is MetricType.CUMULATIVE:
                    linkable_sets_for_measure.append(
                        self._get_linkable_element_set_for_measure(
                            measure,
                            LinkableElementFilter(without_any_of=frozenset({LinkableElementProperty.DATE_PART})),
                        )
                    )
                elif (
                    metric.type is MetricType.SIMPLE
                    or metric.type is MetricType.DERIVED
                    or metric.type is MetricType.RATIO
                ):
                    linkable_sets_for_measure.append(self._get_linkable_element_set_for_measure(measure))
                elif metric.type is MetricType.CONVERSION:
                    conversion_type_params = metric.type_params.conversion_type_params
                    assert (
                        conversion_type_params
                    ), "A conversion metric should have type_params.conversion_type_params defined."
                    if measure == conversion_type_params.base_measure.measure_reference:
                        # Only can query against the base measure's linkable elements
                        # as it joins everything back to the base measure data set so
                        # there is no way of getting the conversion elements
                        linkable_sets_for_measure.append(self._get_linkable_element_set_for_measure(measure))
                else:
                    assert_values_exhausted(metric.type)

            self._metric_to_linkable_element_sets[metric.name] = linkable_sets_for_measure

        # Populate storage dicts with linkable metrics. This loop must happen after the one above so that
        # _metric_to_linkable_element_sets is populated with entities and dimensions.
        linkable_metrics_start_time = time.time()
        for metric in self._semantic_manifest.metrics:
            # Cumulative metrics and time offset metrics require grouping by metric_time, which is not yet available for
            # linkable metrics. So skip those.
            if self._metric_requires_metric_time(metric):
                continue
            metric_reference = MetricReference(metric.name)
            linkable_element_set_for_metric = self.get_linkable_elements_for_metrics([metric_reference])

            defined_from_semantic_models = tuple(
                self._semantic_model_lookup.measure_lookup.get_properties(
                    input_measure.measure_reference
                ).model_reference
                for input_measure in metric.input_measures
            )

            for linkable_entities in linkable_element_set_for_metric.path_key_to_linkable_entities.values():
                for linkable_entity in linkable_entities:
                    # TODO: some users encounter a situation in which the entity reference is in the entity links. Debug why.
                    if linkable_entity.reference in linkable_entity.entity_links:
                        logger.debug(
                            LazyFormat(
                                lambda: f"Found entity reference in entity links for linkable entity: {linkable_entity}"
                            )
                        )
                        continue
                    metric_subquery_join_path_element = MetricSubqueryJoinPathElement(
                        metric_reference=metric_reference,
                        derived_from_semantic_models=defined_from_semantic_models,
                        join_on_entity=linkable_entity.reference,
                        entity_links=linkable_entity.entity_links,
                        metric_to_entity_join_path=(linkable_entity.join_path if linkable_entity.join_path else None),
                    )
                    self._joinable_metrics_for_entities[linkable_entity.reference].add(
                        metric_subquery_join_path_element
                    )
                    # TODO: update _metric_to_linkable_element_sets to have linkable metrics

        # If no metrics are specified, the query interface supports querying distinct values for dimensions, entities,
        # and group by metrics.
        linkable_element_sets_for_no_metrics_queries: List[LinkableElementSet] = []
        # Temp: separate loop to track latency for linkable metrics. TODO: combine with loop below.
        for semantic_model in semantic_manifest.semantic_models:
            linkable_element_sets_for_no_metrics_queries.append(
                self.get_joinable_metrics_for_semantic_model(
                    semantic_model, SemanticModelJoinPath(left_semantic_model_reference=semantic_model.reference)
                )
            )
        logger.debug(
            LazyFormat(
                lambda: f"Building valid linkable metrics took: {time.time() - linkable_metrics_start_time:.2f}s"
            )
        )

        for semantic_model in semantic_manifest.semantic_models:
            linkable_element_sets_for_no_metrics_queries.append(self._get_elements_in_semantic_model(semantic_model))

        metric_time_elements_for_no_metrics = self._get_metric_time_elements(measure_reference=None)
        self._no_metric_linkable_element_set = LinkableElementSet.merge_by_path_key(
            linkable_element_sets_for_no_metrics_queries + [metric_time_elements_for_no_metrics]
        )

        logger.debug(LazyFormat(lambda: f"Building valid group-by-item indexes took: {time.time() - start_time:.2f}s"))

    def _is_semantic_model_scd(self, semantic_model: SemanticModel) -> bool:
        """Whether the semantic model's underlying table is an SCD."""
        return any(dim.validity_params is not None for dim in semantic_model.dimensions)

    def _generate_linkable_time_dimensions(
        self,
        semantic_model_origin: SemanticModelReference,
        dimension: Dimension,
        entity_links: Tuple[EntityReference, ...],
        join_path: SemanticModelJoinPath,
        with_properties: FrozenSet[LinkableElementProperty],
    ) -> Sequence[LinkableDimension]:
        """Generates different versions of the given time dimension with all compatible time granularities & date parts."""
        defined_time_granularity = (
            dimension.type_params.time_granularity if dimension.type_params else DEFAULT_TIME_GRANULARITY
        )

        def create(
            time_granularity: ExpandedTimeGranularity, date_part: Optional[DatePart] = None
        ) -> LinkableDimension:
            properties = set(with_properties)
            if time_granularity.is_custom_granularity or time_granularity.base_granularity != defined_time_granularity:
                properties.add(LinkableElementProperty.DERIVED_TIME_GRANULARITY)
            if date_part:
                properties.add(LinkableElementProperty.DATE_PART)
            return LinkableDimension.create(
                defined_in_semantic_model=semantic_model_origin,
                element_name=dimension.reference.element_name,
                dimension_type=DimensionType.TIME,
                entity_links=entity_links,
                join_path=join_path,
                time_granularity=time_granularity,
                date_part=date_part,
                properties=tuple(sorted(properties)),
            )

        # Build linkable dimensions for all compatible standard granularities and date parts.
        linkable_dimensions = []
        for time_granularity in TimeGranularity:
            if time_granularity.to_int() < defined_time_granularity.to_int():
                continue
            expanded_granularity = ExpandedTimeGranularity.from_time_granularity(time_granularity)
            linkable_dimensions.append(create(expanded_granularity))
            for date_part in DatePart:
                if time_granularity.to_int() <= date_part.to_int():
                    linkable_dimensions.append(create(time_granularity=expanded_granularity, date_part=date_part))

        # Build linkable dimensions for all compatible custom granularities.
        for custom_grain in self._custom_granularities.values():
            if custom_grain.base_granularity.to_int() >= defined_time_granularity.to_int():
                linkable_dimensions.append(create(custom_grain))

        return linkable_dimensions

    def _metric_requires_metric_time(self, metric: Metric) -> bool:
        """Checks if the metric can only be queried with metric_time. Also checks input metrics.

        True if the metric uses cumulative time component or a time offset.
        """
        metrics_to_check = [metric]
        while metrics_to_check:
            metric_to_check = metrics_to_check.pop()
            if metric_to_check.type_params.cumulative_type_params and (
                metric_to_check.type_params.cumulative_type_params.window is not None
                or metric_to_check.type_params.cumulative_type_params.grain_to_date is not None
            ):
                return True
            for input_metric in metric_to_check.input_metrics:
                if input_metric.offset_window is not None or input_metric.offset_to_grain is not None:
                    return True
                metric_for_input_metric = self._metric_references_to_metrics.get(MetricReference(input_metric.name))
                assert (
                    metric_for_input_metric
                ), f"Did not find input metric {input_metric.name} in registered metrics. This indicates internal misconfiguration."
                metrics_to_check.append(metric_for_input_metric)

        return False

    def _get_semantic_model_for_measure(self, measure_reference: MeasureReference) -> SemanticModel:
        semantic_models_where_measure_was_found = []
        for semantic_model in self._semantic_models:
            if any([x.reference.element_name == measure_reference.element_name for x in semantic_model.measures]):
                semantic_models_where_measure_was_found.append(semantic_model)

        if len(semantic_models_where_measure_was_found) == 0:
            raise ValueError(f"No semantic models were found with {measure_reference} in the model")
        elif len(semantic_models_where_measure_was_found) > 1:
            raise ValueError(
                f"Measure {measure_reference} was found in multiple semantic models:\n"
                f"{mf_pformat(semantic_models_where_measure_was_found)}"
            )
        return semantic_models_where_measure_was_found[0]

    def get_joinable_metrics_for_semantic_model(
        self, semantic_model: SemanticModel, using_join_path: SemanticModelJoinPath
    ) -> LinkableElementSet:
        """Get the set of linkable metrics that can be joined to this semantic model.

        From @courtneyholcomb re: `using_join_path`:
            That would be the join path to get from the outer query to the metric subquery. If you query metric1
            filtered by metric2 (grouped by entity1), it would be the join path to get from metric1 to entity1. If
            entity1 is local to the semantic model where metric1's input measures are defined, no join path is
            necessary.
        """
        properties = frozenset({LinkableElementProperty.METRIC, LinkableElementProperty.JOINED})
        if self._is_semantic_model_scd(semantic_model):
            properties = properties.union({LinkableElementProperty.SCD_HOP})

        join_path_has_path_links = len(using_join_path.path_elements) > 0
        if join_path_has_path_links:
            assert semantic_model.reference == using_join_path.last_semantic_model_reference, (
                "Last join path element should match semantic model when building LinkableMetrics. "
                f"Got semantic model: {semantic_model.reference.semantic_model_name}; "
                f"last join path element: {using_join_path.last_semantic_model_reference.semantic_model_name}",
            )
            # Temp: disable LinkableMetrics with outer join path until there is an interface to specify it.
            # The result's properties = properties.union(frozenset({LinkableElementProperty.MULTI_HOP}))
            return LinkableElementSet()

        path_key_to_linkable_metrics: Dict[ElementPathKey, Tuple[LinkableMetric, ...]] = {}
        for entity_reference in [entity.reference for entity in semantic_model.entities]:
            # Avoid creating an entity link cycle.
            if join_path_has_path_links and entity_reference in using_join_path.entity_links:
                continue
            for metric_subquery_join_path_element in self._joinable_metrics_for_entities[entity_reference]:
                linkable_metric = LinkableMetric.create(
                    properties=properties,
                    join_path=SemanticModelToMetricSubqueryJoinPath(
                        metric_subquery_join_path_element=metric_subquery_join_path_element,
                        semantic_model_join_path=using_join_path,
                    ),
                )
                path_key_to_linkable_metrics[linkable_metric.path_key] = path_key_to_linkable_metrics.get(
                    linkable_metric.path_key, ()
                ) + (linkable_metric,)

        return LinkableElementSet(path_key_to_linkable_metrics=path_key_to_linkable_metrics)

    def _get_elements_in_semantic_model(self, semantic_model: SemanticModel) -> LinkableElementSet:
        """Gets the elements in the semantic model, without requiring any joins.

        Elements related to metric_time are handled separately in _get_metric_time_elements().
        Linkable metrics are not considered local to the semantic model since they always require a join.
        """
        semantic_model_is_scd = self._is_semantic_model_scd(semantic_model)

        linkable_dimensions = []
        linkable_entities = []

        entity_properties = frozenset({LinkableElementProperty.LOCAL, LinkableElementProperty.ENTITY})
        if semantic_model_is_scd:
            entity_properties = entity_properties.union({LinkableElementProperty.SCD_HOP})

        for entity in semantic_model.entities:
            linkable_entities.append(
                LinkableEntity.create(
                    defined_in_semantic_model=semantic_model.reference,
                    element_name=entity.reference.element_name,
                    entity_links=(),
                    join_path=SemanticModelJoinPath(
                        left_semantic_model_reference=semantic_model.reference,
                    ),
                    properties=entity_properties,
                )
            )
            for entity_link in SemanticModelHelper.entity_links_for_local_elements(semantic_model):
                # Avoid creating "booking_id__booking_id"
                if entity_link == entity.reference:
                    continue
                linkable_entities.append(
                    LinkableEntity.create(
                        defined_in_semantic_model=semantic_model.reference,
                        element_name=entity.reference.element_name,
                        entity_links=(entity_link,),
                        join_path=SemanticModelJoinPath(
                            left_semantic_model_reference=semantic_model.reference,
                        ),
                        properties=entity_properties,
                    )
                )

        dimension_properties = frozenset({LinkableElementProperty.LOCAL})
        if semantic_model_is_scd:
            dimension_properties = dimension_properties.union({LinkableElementProperty.SCD_HOP})

        for entity_link in SemanticModelHelper.entity_links_for_local_elements(semantic_model):
            for dimension in semantic_model.dimensions:
                dimension_type = dimension.type
                if dimension_type is DimensionType.CATEGORICAL:
                    linkable_dimensions.append(
                        LinkableDimension.create(
                            defined_in_semantic_model=semantic_model.reference,
                            element_name=dimension.reference.element_name,
                            dimension_type=DimensionType.CATEGORICAL,
                            entity_links=(entity_link,),
                            join_path=SemanticModelJoinPath(
                                left_semantic_model_reference=semantic_model.reference,
                            ),
                            properties=dimension_properties,
                            time_granularity=None,
                            date_part=None,
                        )
                    )
                elif dimension_type is DimensionType.TIME:
                    linkable_dimensions.extend(
                        self._generate_linkable_time_dimensions(
                            semantic_model_origin=semantic_model.reference,
                            dimension=dimension,
                            entity_links=(entity_link,),
                            join_path=SemanticModelJoinPath(
                                left_semantic_model_reference=semantic_model.reference,
                            ),
                            with_properties=dimension_properties,
                        )
                    )
                else:
                    assert_values_exhausted(dimension_type)

        path_key_to_linkable_dimensions: Dict[ElementPathKey, Tuple[LinkableDimension, ...]] = {}
        for linkable_dimension in linkable_dimensions:
            path_key_to_linkable_dimensions[linkable_dimension.path_key] = path_key_to_linkable_dimensions.get(
                linkable_dimension.path_key, ()
            ) + (linkable_dimension,)
        path_key_to_linkable_entities: Dict[ElementPathKey, Tuple[LinkableEntity, ...]] = {}
        for linkable_entity in linkable_entities:
            path_key_to_linkable_entities[linkable_entity.path_key] = path_key_to_linkable_entities.get(
                linkable_entity.path_key, ()
            ) + (linkable_entity,)
        return LinkableElementSet(
            path_key_to_linkable_dimensions=path_key_to_linkable_dimensions,
            path_key_to_linkable_entities=path_key_to_linkable_entities,
        )

    def _get_semantic_models_with_joinable_entity(
        self,
        left_semantic_model_reference: SemanticModelReference,
        entity_reference: EntityReference,
    ) -> Sequence[SemanticModel]:
        # May switch to non-cached implementation.
        semantic_models = self._entity_to_semantic_model[entity_reference.element_name]
        valid_semantic_models = []
        for semantic_model in semantic_models:
            if self._join_evaluator.is_valid_semantic_model_join(
                left_semantic_model_reference=left_semantic_model_reference,
                right_semantic_model_reference=semantic_model.reference,
                on_entity_reference=entity_reference,
            ):
                valid_semantic_models.append(semantic_model)
        return valid_semantic_models

    @staticmethod
    def _get_time_granularity_for_dimension(
        semantic_model: SemanticModel, time_dimension_reference: TimeDimensionReference
    ) -> TimeGranularity:
        matching_dimensions = tuple(
            dimension
            for dimension in semantic_model.dimensions
            if dimension.type is DimensionType.TIME and dimension.time_dimension_reference == time_dimension_reference
        )

        if len(matching_dimensions) != 1:
            raise RuntimeError(
                f"Did not find a matching time dimension for {time_dimension_reference} in {semantic_model}"
            )
        time_dimension = matching_dimensions[0]
        type_params = time_dimension.type_params
        assert (
            type_params is not None
        ), f"type_params should have been set for {time_dimension_reference} in {semantic_model}"

        assert (
            type_params.time_granularity is not None
        ), f"time_granularity should have been set for {time_dimension_reference} in {semantic_model}"

        return type_params.time_granularity

    def _get_metric_time_elements(self, measure_reference: Optional[MeasureReference] = None) -> LinkableElementSet:
        """Create elements for metric_time for a given measure in a semantic model.

        metric_time is a virtual dimension that is the same as aggregation time dimension for a measure, but with a
        different name. Because it doesn't actually exist in the semantic model, these elements need to be created based
        on what aggregation time dimension was used to define the measure.
        """
        measure_semantic_model: Optional[SemanticModel] = None
        defined_granularity: Optional[ExpandedTimeGranularity] = None
        if measure_reference:
            measure_semantic_model = self._get_semantic_model_for_measure(measure_reference)
            semantic_model_is_scd = self._is_semantic_model_scd(measure_semantic_model)
            measure_agg_time_dimension_reference = measure_semantic_model.checked_agg_time_dimension_for_measure(
                measure_reference=measure_reference
            )
            min_granularity = self._get_time_granularity_for_dimension(
                semantic_model=measure_semantic_model,
                time_dimension_reference=measure_agg_time_dimension_reference,
            )
            defined_granularity = ExpandedTimeGranularity.from_time_granularity(min_granularity)
        else:
            # If querying metric_time without metrics, will query from time spines.
            # Defaults to DAY granularity if available in time spines, else smallest available granularity.
            min_granularity = min(self._time_spine_sources.keys())
            semantic_model_is_scd = False
        possible_metric_time_granularities = tuple(
            ExpandedTimeGranularity.from_time_granularity(time_granularity)
            for time_granularity in TimeGranularity
            if min_granularity.is_smaller_than_or_equal(time_granularity)
        ) + tuple(
            [
                custom_granularity
                for custom_granularity in self._custom_granularities.values()
                if min_granularity.is_smaller_than_or_equal(custom_granularity.base_granularity)
            ]
        )

        # For each granularity, will create one LinkableDimension with no date part, and one for each compatible date part.
        # TODO: group by resolution has different logic than source node builder for combining date part w/ grain. Fix.
        path_key_to_linkable_dimensions: Dict[ElementPathKey, List[LinkableDimension]] = defaultdict(list)
        for time_granularity in possible_metric_time_granularities:
            possible_date_parts: Tuple[Optional[DatePart], ...] = (None,)
            if not time_granularity.is_custom_granularity:
                possible_date_parts += tuple(
                    date_part
                    for date_part in DatePart
                    if time_granularity.base_granularity.to_int() <= date_part.to_int()
                )

            for date_part in possible_date_parts:
                properties = {LinkableElementProperty.METRIC_TIME}
                if time_granularity != defined_granularity:
                    properties.add(LinkableElementProperty.DERIVED_TIME_GRANULARITY)
                if date_part:
                    properties.add(LinkableElementProperty.DATE_PART)
                if semantic_model_is_scd:
                    properties.add(LinkableElementProperty.SCD_HOP)
                linkable_dimension = LinkableDimension.create(
                    defined_in_semantic_model=measure_semantic_model.reference if measure_semantic_model else None,
                    element_name=MetricFlowReservedKeywords.METRIC_TIME.value,
                    dimension_type=DimensionType.TIME,
                    entity_links=(),
                    join_path=SemanticModelJoinPath(
                        left_semantic_model_reference=(
                            measure_semantic_model.reference
                            if measure_semantic_model
                            else SemanticModelDerivation.VIRTUAL_SEMANTIC_MODEL_REFERENCE
                        ),
                    ),
                    properties=frozenset(properties),
                    time_granularity=time_granularity,
                    date_part=date_part,
                )
                path_key_to_linkable_dimensions[linkable_dimension.path_key].append(linkable_dimension)

        return LinkableElementSet(
            path_key_to_linkable_dimensions={
                path_key: tuple(linkable_dimensions)
                for path_key, linkable_dimensions in path_key_to_linkable_dimensions.items()
            }
        )

    def _get_joined_elements(self, measure_semantic_model_reference: SemanticModelReference) -> LinkableElementSet:
        """Get the elements that can be generated by joining other models to the given model."""
        result = self._semantic_model_reference_to_joined_elements.get(measure_semantic_model_reference)
        if result is not None:
            return result

        result = self._get_joined_elements_without_cache(measure_semantic_model_reference)
        self._semantic_model_reference_to_joined_elements[measure_semantic_model_reference] = result
        return result

    def _get_joined_elements_without_cache(
        self, measure_semantic_model_reference: SemanticModelReference
    ) -> LinkableElementSet:
        measure_semantic_model = self._semantic_model_reference_to_semantic_model[measure_semantic_model_reference]
        # Create single-hop elements
        join_paths = []
        for entity in measure_semantic_model.entities:
            semantic_models = self._get_semantic_models_with_joinable_entity(
                left_semantic_model_reference=measure_semantic_model.reference,
                entity_reference=entity.reference,
            )
            for semantic_model in semantic_models:
                if semantic_model.name == measure_semantic_model.name:
                    continue
                join_paths.append(
                    SemanticModelJoinPath.from_single_element(
                        left_semantic_model_reference=measure_semantic_model.reference,
                        right_semantic_model_reference=semantic_model.reference,
                        join_on_entity=entity.reference,
                    )
                )
        single_hop_elements = LinkableElementSet.merge_by_path_key(
            [self.create_linkable_element_set_from_join_path(join_path) for join_path in join_paths]
        )

        # Create multi-hop elements. At each iteration, we generate the list of valid elements based on the current join
        # path, extend all paths to include the next valid semantic model, then repeat.
        multi_hop_elements = LinkableElementSet()

        for _ in range(self._max_entity_links - 1):
            new_join_paths: List[SemanticModelJoinPath] = []
            for join_path in join_paths:
                new_join_paths.extend(
                    self._find_next_possible_paths(
                        measure_semantic_model=measure_semantic_model, current_join_path=join_path
                    )
                )

            if len(new_join_paths) == 0:
                break

            multi_hop_elements = LinkableElementSet.merge_by_path_key(
                (multi_hop_elements,)
                + tuple(
                    self.create_linkable_element_set_from_join_path(new_join_path) for new_join_path in new_join_paths
                )
            )
            join_paths = new_join_paths

        return LinkableElementSet.merge_by_path_key((single_hop_elements, multi_hop_elements))

    def _get_linkable_element_set_for_measure(
        self,
        measure_reference: MeasureReference,
        element_filter: LinkableElementFilter = LinkableElementFilter(),
    ) -> LinkableElementSet:
        """See get_linkable_element_set_for_measure()."""
        measure_semantic_model = self._get_semantic_model_for_measure(measure_reference)

        elements_in_semantic_model = self._get_elements_in_semantic_model(measure_semantic_model)

        # Filter out group-by metrics if not specified by the property as there can be a large number of them.
        if LinkableElementProperty.METRIC not in element_filter.without_any_of:
            metrics_linked_to_semantic_model = self.get_joinable_metrics_for_semantic_model(
                semantic_model=measure_semantic_model,
                using_join_path=SemanticModelJoinPath(left_semantic_model_reference=measure_semantic_model.reference),
            )
        else:
            metrics_linked_to_semantic_model = LinkableElementSet()

        metric_time_elements = self._get_metric_time_elements(measure_reference)
        joined_elements = self._get_joined_elements(measure_semantic_model.reference)

        return LinkableElementSet.merge_by_path_key(
            (
                elements_in_semantic_model,
                metrics_linked_to_semantic_model,
                metric_time_elements,
                joined_elements,
            )
        ).filter(element_filter)

    def get_linkable_element_set_for_measure(
        self,
        measure_reference: MeasureReference,
        element_filter: LinkableElementFilter,
    ) -> LinkableElementSet:
        """Get the valid linkable elements for the given measure."""
        return self._get_linkable_element_set_for_measure(
            measure_reference=measure_reference,
            element_filter=element_filter,
        )

    def get_linkable_elements_for_distinct_values_query(
        self,
        element_filter: LinkableElementFilter,
    ) -> LinkableElementSet:
        """Returns queryable items for a distinct group-by-item values query.

        A distinct group-by-item values query does not include any metrics.
        """
        return self._no_metric_linkable_element_set.filter(element_filter)

    # TODO: the results of this method don't actually match what will be allowed for the metric. This method checks
    # _metric_to_linkable_element_sets, while the actual group by resolution DAG calls _get_linkable_element_set_for_measure.
    def get_linkable_elements_for_metrics(
        self,
        metric_references: Sequence[MetricReference],
        element_filter: LinkableElementFilter = LinkableElementFilter(),
    ) -> LinkableElementSet:
        """Gets the valid linkable elements that are common to all requested metrics."""
        linkable_element_sets = []
        for metric_reference in metric_references:
            element_sets = self._metric_to_linkable_element_sets.get(metric_reference.element_name)
            if not element_sets:
                raise UnknownMetricLinkingError(f"Unknown metric: {metric_reference} in element set")

            # Using .only_unique_path_keys to exclude ambiguous elements where there are multiple join paths to get
            # a dimension / entity.
            metric_result = LinkableElementSet.intersection_by_path_key(
                [element_set.only_unique_path_keys.filter(element_filter) for element_set in element_sets]
            )
            linkable_element_sets.append(metric_result)

        intersection_set = LinkableElementSet.intersection_by_path_key(linkable_element_sets)
        return intersection_set

    def _find_next_possible_paths(
        self, measure_semantic_model: SemanticModel, current_join_path: SemanticModelJoinPath
    ) -> Sequence[SemanticModelJoinPath]:
        """Generate the set of possible paths that are 1 semantic model join longer that the "current_join_path"."""
        last_semantic_model_in_path = self._semantic_model_lookup.get_by_reference(
            current_join_path.last_semantic_model_reference
        )
        assert last_semantic_model_in_path
        new_join_paths = []

        for entity in last_semantic_model_in_path.entities:
            entity_reference = entity.reference

            # Don't create cycles in the join path by joining on the same entity.
            if entity_reference in set(x.join_on_entity for x in current_join_path.path_elements):
                continue

            semantic_models_that_can_be_joined = self._get_semantic_models_with_joinable_entity(
                left_semantic_model_reference=last_semantic_model_in_path.reference,
                entity_reference=entity.reference,
            )
            for semantic_model in semantic_models_that_can_be_joined:
                # Don't create cycles in the join path by repeating a semantic model in the path.
                if semantic_model.name == measure_semantic_model.name or any(
                    tuple(
                        path_element.semantic_model_reference == semantic_model.reference
                        for path_element in current_join_path.path_elements
                    )
                ):
                    continue

                new_join_path = SemanticModelJoinPath(
                    left_semantic_model_reference=current_join_path.left_semantic_model_reference,
                    path_elements=current_join_path.path_elements
                    + (
                        SemanticModelJoinPathElement(
                            semantic_model_reference=semantic_model.reference, join_on_entity=entity_reference
                        ),
                    ),
                )
                new_join_paths.append(new_join_path)

        return new_join_paths

    def create_linkable_element_set_from_join_path(
        self,
        join_path: SemanticModelJoinPath,
    ) -> LinkableElementSet:
        """Given the current path, generate the respective linkable elements from the last semantic model in the path."""
        semantic_model = self._semantic_model_lookup.get_by_reference(join_path.last_semantic_model_reference)
        assert (
            semantic_model
        ), f"Semantic model {join_path.last_semantic_model_reference.semantic_model_name} is in join path but does not exist in SemanticModelLookup"

        properties = frozenset({LinkableElementProperty.JOINED})
        if len(join_path.path_elements) > 1:
            properties = properties.union({LinkableElementProperty.MULTI_HOP})

        # If any of the semantic models in the join path is an SCD, add SCD_HOP
        for reference_to_derived_model in join_path.derived_from_semantic_models:
            derived_model = self._semantic_model_lookup.get_by_reference(reference_to_derived_model)
            assert (
                derived_model
            ), f"Semantic model {reference_to_derived_model.semantic_model_name} is in join path but does not exist in SemanticModelLookup"

            if self._is_semantic_model_scd(derived_model):
                properties = properties.union({LinkableElementProperty.SCD_HOP})
                break

        linkable_dimensions: List[LinkableDimension] = []
        linkable_entities: List[LinkableEntity] = []

        for dimension in semantic_model.dimensions:
            dimension_type = dimension.type
            if dimension_type == DimensionType.CATEGORICAL:
                linkable_dimensions.append(
                    LinkableDimension.create(
                        defined_in_semantic_model=semantic_model.reference,
                        element_name=dimension.reference.element_name,
                        dimension_type=DimensionType.CATEGORICAL,
                        entity_links=join_path.entity_links,
                        join_path=join_path,
                        properties=properties,
                        time_granularity=None,
                        date_part=None,
                    )
                )
            elif dimension_type == DimensionType.TIME:
                linkable_dimensions.extend(
                    self._generate_linkable_time_dimensions(
                        semantic_model_origin=semantic_model.reference,
                        dimension=dimension,
                        entity_links=join_path.entity_links,
                        join_path=join_path,
                        with_properties=properties,
                    )
                )
            else:
                raise RuntimeError(f"Unhandled type: {dimension_type}")

        for entity in semantic_model.entities:
            # Avoid creating "booking_id__booking_id"
            if entity.reference != join_path.last_entity_link:
                linkable_entities.append(
                    LinkableEntity.create(
                        defined_in_semantic_model=semantic_model.reference,
                        element_name=entity.reference.element_name,
                        entity_links=join_path.entity_links,
                        join_path=join_path,
                        properties=properties.union({LinkableElementProperty.ENTITY}),
                    )
                )

        path_key_to_linkable_dimensions: Dict[ElementPathKey, Tuple[LinkableDimension, ...]] = {}
        for linkable_dimension in linkable_dimensions:
            path_key_to_linkable_dimensions[linkable_dimension.path_key] = path_key_to_linkable_dimensions.get(
                linkable_dimension.path_key, ()
            ) + (linkable_dimension,)
        path_key_to_linkable_entities: Dict[ElementPathKey, Tuple[LinkableEntity, ...]] = {}
        for linkable_entity in linkable_entities:
            path_key_to_linkable_entities[linkable_entity.path_key] = path_key_to_linkable_entities.get(
                linkable_entity.path_key, ()
            ) + (linkable_entity,)
        return LinkableElementSet(
            path_key_to_linkable_dimensions=path_key_to_linkable_dimensions,
            path_key_to_linkable_entities=path_key_to_linkable_entities,
            path_key_to_linkable_metrics=self.get_joinable_metrics_for_semantic_model(
                semantic_model=semantic_model, using_join_path=join_path
            ).path_key_to_linkable_metrics,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Optional

from dbt_semantic_interfaces.protocols.entity import EntityType
from dbt_semantic_interfaces.references import (
    EntityReference,
    SemanticModelElementReference,
    SemanticModelReference,
)

if TYPE_CHECKING:
    from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup

MAX_JOIN_HOPS = 2


@dataclass(frozen=True)
class SemanticModelEntityJoinType:
    """Describe a type of join between semantic models where entities are of the listed types."""

    left_entity_type: EntityType
    right_entity_type: EntityType


@dataclass(frozen=True)
class SemanticModelEntityJoin:
    """How to join one semantic model onto another, using a specific entity and join type."""

    right_semantic_model_reference: SemanticModelReference
    entity_reference: EntityReference
    join_type: SemanticModelEntityJoinType


@dataclass(frozen=True)
class SemanticModelLink:
    """The valid join path to link two semantic models. Might include multiple joins."""

    left_semantic_model_reference: SemanticModelReference
    join_path: List[SemanticModelEntityJoin]


class SemanticModelJoinEvaluator:
    """Checks to see if a join between two semantic models should be allowed."""

    # Valid joins are the non-fanout joins.
    _VALID_ENTITY_JOINS = (
        SemanticModelEntityJoinType(left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.NATURAL),
        SemanticModelEntityJoinType(left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.PRIMARY),
        SemanticModelEntityJoinType(left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.UNIQUE),
        SemanticModelEntityJoinType(left_entity_type=EntityType.UNIQUE, right_entity_type=EntityType.NATURAL),
        SemanticModelEntityJoinType(left_entity_type=EntityType.UNIQUE, right_entity_type=EntityType.PRIMARY),
        SemanticModelEntityJoinType(left_entity_type=EntityType.UNIQUE, right_entity_type=EntityType.UNIQUE),
        SemanticModelEntityJoinType(left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.NATURAL),
        SemanticModelEntityJoinType(left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.PRIMARY),
        SemanticModelEntityJoinType(left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.UNIQUE),
        SemanticModelEntityJoinType(left_entity_type=EntityType.NATURAL, right_entity_type=EntityType.PRIMARY),
        SemanticModelEntityJoinType(left_entity_type=EntityType.NATURAL, right_entity_type=EntityType.UNIQUE),
    )

    _INVALID_ENTITY_JOINS = (
        SemanticModelEntityJoinType(left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.FOREIGN),
        SemanticModelEntityJoinType(left_entity_type=EntityType.UNIQUE, right_entity_type=EntityType.FOREIGN),
        SemanticModelEntityJoinType(left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.FOREIGN),
        SemanticModelEntityJoinType(left_entity_type=EntityType.NATURAL, right_entity_type=EntityType.FOREIGN),
        # Natural -> Natural joins are not allowed due to hidden fanout or missing value concerns with
        # multiple validity windows in play
        SemanticModelEntityJoinType(left_entity_type=EntityType.NATURAL, right_entity_type=EntityType.NATURAL),
    )

    def __init__(self, semantic_model_lookup: SemanticModelLookup) -> None:  # noqa: D107
        self._semantic_model_lookup = semantic_model_lookup

    def get_joinable_semantic_models(
        self, left_semantic_model_reference: SemanticModelReference, include_multi_hop: bool = False
    ) -> Dict[str, SemanticModelLink]:
        """List all semantic models that can join to given semantic model, and the entities to join them."""
        semantic_model_joins: Dict[str, SemanticModelLink] = {}
        self._get_remaining_hops_of_joinable_semantic_models(
            left_semantic_model_reference=left_semantic_model_reference,
            parent_semantic_model_to_join_paths={left_semantic_model_reference: []},
            known_semantic_model_joins=semantic_model_joins,
            join_hops_remaining=(MAX_JOIN_HOPS if include_multi_hop else 1),
        )
        return semantic_model_joins

    def _get_remaining_hops_of_joinable_semantic_models(
        self,
        left_semantic_model_reference: SemanticModelReference,
        parent_semantic_model_to_join_paths: Dict[SemanticModelReference, List[SemanticModelEntityJoin]],
        known_semantic_model_joins: Dict[str, SemanticModelLink],
        join_hops_remaining: int,
    ) -> None:
        assert join_hops_remaining > 0, "No join hops remaining. This is unexpected with proper use of this method."
        for parent_semantic_model_reference, parent_join_path in parent_semantic_model_to_join_paths.items():
            parent_semantic_model = self._semantic_model_lookup.get_by_reference(
                semantic_model_reference=parent_semantic_model_reference
            )
            assert parent_semantic_model is not None

            # We'll get all joinable semantic models in this hop before recursing to ensure we find the most
            # efficient path to each semantic model.
            join_paths_to_visit_next: List[List[SemanticModelEntityJoin]] = []
            for entity in parent_semantic_model.entities:
                entity_reference = EntityReference(element_name=entity.name)
                entity_semantic_models = self._semantic_model_lookup.get_semantic_models_for_entity(
                    entity_reference=entity_reference
                )

                for right_semantic_model in entity_semantic_models:
                    # Check if we've seen this semantic model already
                    if (
                        right_semantic_model.name == left_semantic_model_reference.semantic_model_name
                        or right_semantic_model.name in known_semantic_model_joins
                    ):
                        continue

                    # Check if there is a valid way to join this semantic model to existing join path
                    right_semantic_model_reference = SemanticModelReference(
                        semantic_model_name=right_semantic_model.name
                    )
                    valid_join_type = self.get_valid_semantic_model_entity_join_type(
                        left_semantic_model_reference=parent_semantic_model_reference,
                        right_semantic_model_reference=right_semantic_model_reference,
                        on_entity_reference=entity_reference,
                    )
                    if valid_join_type is None:
                        continue

                    join_path_for_semantic_model = parent_join_path + [
                        SemanticModelEntityJoin(
                            right_semantic_model_reference=right_semantic_model_reference,
                            entity_reference=entity_reference,
                            join_type=valid_join_type,
                        )
                    ]
                    join_paths_to_visit_next.append(join_path_for_semantic_model)
                    known_semantic_model_joins[right_semantic_model_reference.semantic_model_name] = SemanticModelLink(
                        left_semantic_model_reference=left_semantic_model_reference,
                        join_path=join_path_for_semantic_model,
                    )

        join_hops_remaining -= 1
        if not join_hops_remaining:
            return

        right_semantic_models_to_join_paths: Dict[SemanticModelReference, List[SemanticModelEntityJoin]] = {}
        for join_path in join_paths_to_visit_next:
            assert len(join_path) > 0
            right_semantic_models_to_join_paths[join_path[-1].right_semantic_model_reference] = join_path

        self._get_remaining_hops_of_joinable_semantic_models(
            left_semantic_model_reference=left_semantic_model_reference,
            parent_semantic_model_to_join_paths=right_semantic_models_to_join_paths,
            known_semantic_model_joins=known_semantic_model_joins,
            join_hops_remaining=join_hops_remaining,
        )

    def get_valid_semantic_model_entity_join_type(
        self,
        left_semantic_model_reference: SemanticModelReference,
        right_semantic_model_reference: SemanticModelReference,
        on_entity_reference: EntityReference,
    ) -> Optional[SemanticModelEntityJoinType]:
        """Get valid join type used to join semantic models on given entity, if exists."""
        left_entity = self._semantic_model_lookup.get_entity_in_semantic_model(
            SemanticModelElementReference.create_from_references(left_semantic_model_reference, on_entity_reference)
        )

        right_entity = self._semantic_model_lookup.get_entity_in_semantic_model(
            SemanticModelElementReference.create_from_references(right_semantic_model_reference, on_entity_reference)
        )
        if left_entity is None or right_entity is None:
            return None

        left_semantic_model = self._semantic_model_lookup.get_by_reference(left_semantic_model_reference)
        right_semantic_model = self._semantic_model_lookup.get_by_reference(right_semantic_model_reference)
        assert left_semantic_model, "Type refinement. If you see this error something has refactored wrongly"
        assert right_semantic_model, "Type refinement. If you see this error something has refactored wrongly"

        if left_semantic_model.has_validity_dimensions and right_semantic_model.has_validity_dimensions:
            # We cannot join two semantic models with validity dimensions due to concerns with unexpected fanout
            # due to the key structure of these semantic models. Applying multi-stage validity window filters can
            # also lead to unexpected removal of interim join keys. Note this will need to be updated if we enable
            # measures in such semantic models, since those will need to be converted to a different type of semantic model
            # to support measure computation.
            return None

        if right_entity.type is EntityType.NATURAL:
            if not right_semantic_model.has_validity_dimensions:
                # There is no way to refine this to a single row per key, so we cannot support this join
                return None

        join_type = SemanticModelEntityJoinType(left_entity.type, right_entity.type)

        if join_type in SemanticModelJoinEvaluator._VALID_ENTITY_JOINS:
            return join_type
        elif join_type in SemanticModelJoinEvaluator._INVALID_ENTITY_JOINS:
            return None

        raise RuntimeError(f"Join type not handled: {join_type}")

    def is_valid_semantic_model_join(
        self,
        left_semantic_model_reference: SemanticModelReference,
        right_semantic_model_reference: SemanticModelReference,
        on_entity_reference: EntityReference,
    ) -> bool:
        """Return true if we should allow a join with the given parameters to resolve a query."""
        return (
            self.get_valid_semantic_model_entity_join_type(
                left_semantic_model_reference=left_semantic_model_reference,
                right_semantic_model_reference=right_semantic_model_reference,
                on_entity_reference=on_entity_reference,
            )
            is not None
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional


class SqlTableType(Enum):  # noqa: D101
    TABLE = "table"
    VIEW = "view"
    # CTE type may be added later.


@dataclass(frozen=True, order=True)
class SqlTable:
    """Represents a reference to a SQL table."""

    schema_name: Optional[str]
    table_name: str
    db_name: Optional[str] = None
    table_type: SqlTableType = SqlTableType.TABLE

    def __post_init__(self) -> None:  # noqa: D105
        if self.db_name is not None and self.schema_name is None:
            raise ValueError(f"{self.db_name=} when it should be specified with {self.schema_name=}")

    @staticmethod
    def from_string(sql_str: str) -> SqlTable:  # noqa: D102
        sql_str_split = sql_str.split(".")
        if len(sql_str_split) == 2:
            return SqlTable(schema_name=sql_str_split[0], table_name=sql_str_split[1])
        elif len(sql_str_split) == 3:
            return SqlTable(db_name=sql_str_split[0], schema_name=sql_str_split[1], table_name=sql_str_split[2])
        raise RuntimeError(
            f"Invalid input for a SQL table, expected form '<schema>.<table>' or '<db>.<schema>.<table>' "
            f"but got: {sql_str}"
        )

    @property
    def sql(self) -> str:
        """Return the snippet that can be used for use in SQL queries."""
        items = []
        if self.db_name is not None:
            items.append(self.db_name)
        if self.schema_name is not None:
            items.append(self.schema_name)
        items.append(self.table_name)

        return ".".join(items)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern

if TYPE_CHECKING:
    from metricflow_semantics.specs.instance_spec import InstanceSpec


class QueryItemNamingScheme(ABC):
    """Describes how to name items that are involved in a MetricFlow query.

    Most useful for group-by-items as there are different ways to name them like "user__country"
    or "TimeDimension('metric_time', 'DAY')".
    """

    @abstractmethod
    def input_str(self, instance_spec: InstanceSpec) -> Optional[str]:
        """Following this scheme, return the string that can be used as an input that would specify the given spec.

        This is used to generate suggestions from available group-by-items if the user specifies a group-by-item that is
        invalid.

        If this scheme cannot accommodate the spec, return None. This is needed to handle unsupported cases in
        DunderNamingScheme, such as DatePart, but naming schemes should otherwise be complete.
        """
        pass

    @abstractmethod
    def spec_pattern(self, input_str: str, semantic_manifest_lookup: SemanticManifestLookup) -> SpecPattern:
        """Given an input that follows this scheme, return a spec pattern that matches the described input.

        This is used to generate suggestions from available group-by-items if the user specifies a group-by-item that is
        invalid.

        If this scheme cannot accommodate the spec, return None. This is needed to handle unsupported cases in
        DunderNamingScheme, such as DatePart, but naming schemes should otherwise be complete.
        """
        pass

    @abstractmethod
    def input_str_follows_scheme(self, input_str: str) -> bool:
        """Returns true if the given input string follows this naming scheme.

        Consider adding a structured result that indicates why it does not match the scheme.
        """
        pass



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Optional, Sequence

from dbt_semantic_interfaces.references import SemanticModelReference, TimeDimensionReference
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.validations.unique_valid_name import MetricFlowReservedKeywords
from metricflow_semantics.instances import InstanceSet, TimeDimensionInstance
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


class DataSet(ABC):
    """Describes a set of data that a source node in the dataflow plan contains."""

    def __init__(self, instance_set: InstanceSet) -> None:  # noqa: D107
        self._instance_set = instance_set

    @property
    def instance_set(self) -> InstanceSet:
        """Returns the instances contained in this dataset."""
        return self._instance_set

    @property
    def metric_time_dimension_instances(self) -> Sequence[TimeDimensionInstance]:
        """Extracts all metric time TimeDimensionInstances from the InstanceSet associated with this DataSet."""
        return tuple(
            time_dimension_instance
            for time_dimension_instance in self.instance_set.time_dimension_instances
            if time_dimension_instance.spec.element_name == DataSet.metric_time_dimension_name()
        )

    @staticmethod
    def metric_time_dimension_reference() -> TimeDimensionReference:
        """Returns a special reference that means 'the aggregation time dimension for all measures in the data set'.

        Or to put in another way, if metrics for those measures were plotted together on a graph, this is the name of
        the time dimension for the x-axis.
        """
        return TimeDimensionReference(element_name=MetricFlowReservedKeywords.METRIC_TIME.value)

    @staticmethod
    def metric_time_dimension_name() -> str:
        """See metric_time_dimension_reference(), but in name form for use in queries."""
        return DataSet.metric_time_dimension_reference().element_name

    @staticmethod
    def metric_time_dimension_spec(
        time_granularity: ExpandedTimeGranularity, date_part: Optional[DatePart] = None
    ) -> TimeDimensionSpec:
        """Spec that corresponds to DataSet.metric_time_dimension_reference."""
        return TimeDimensionSpec(
            element_name=DataSet.metric_time_dimension_reference().element_name,
            entity_links=(),
            time_granularity=time_granularity,
            date_part=date_part,
        )

    @property
    @abstractmethod
    def semantic_model_reference(self) -> Optional[SemanticModelReference]:
        """If this data set was created from a semantic model, return the reference."""
        raise NotImplementedError

    def __repr__(self) -> str:  # noqa: D105
        return f"{self.__class__.__name__}()"



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import List, Optional, Sequence

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.specs.metric_spec import MetricSpec

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor
from metricflow.dataflow.nodes.add_generated_uuid import AddGeneratedUuidColumnNode
from metricflow.dataflow.nodes.aggregate_measures import AggregateMeasuresNode
from metricflow.dataflow.nodes.alias_specs import AliasSpecsNode
from metricflow.dataflow.nodes.combine_aggregated_outputs import CombineAggregatedOutputsNode
from metricflow.dataflow.nodes.compute_metrics import ComputeMetricsNode
from metricflow.dataflow.nodes.constrain_time import ConstrainTimeRangeNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataflow.nodes.join_conversion_events import JoinConversionEventsNode
from metricflow.dataflow.nodes.join_over_time import JoinOverTimeRangeNode
from metricflow.dataflow.nodes.join_to_base import JoinOnEntitiesNode
from metricflow.dataflow.nodes.join_to_custom_granularity import JoinToCustomGranularityNode
from metricflow.dataflow.nodes.join_to_time_spine import JoinToTimeSpineNode
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataflow.nodes.min_max import MinMaxNode
from metricflow.dataflow.nodes.order_by_limit import OrderByLimitNode
from metricflow.dataflow.nodes.read_sql_source import ReadSqlSourceNode
from metricflow.dataflow.nodes.semi_additive_join import SemiAdditiveJoinNode
from metricflow.dataflow.nodes.where_filter import WhereConstraintNode
from metricflow.dataflow.nodes.window_reaggregation_node import WindowReaggregationNode
from metricflow.dataflow.nodes.write_to_data_table import WriteToResultDataTableNode
from metricflow.dataflow.nodes.write_to_table import WriteToResultTableNode
from metricflow.dataflow.optimizer.source_scan.matching_linkable_specs import MatchingLinkableSpecsTransform

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ComputeMetricsBranchCombinerResult:  # noqa: D101
    # Perhaps adding more metadata about how nodes got combined would be useful.
    # If combined_branch is None, it means combination could not occur.
    combined_branch: Optional[DataflowPlanNode] = None

    @property
    def combined(self) -> bool:
        """Returns true if this result indicates that the branch could be combined."""
        return self.combined_branch is not None

    @property
    def checked_combined_branch(self) -> DataflowPlanNode:  # noqa: D102
        assert self.combined_branch is not None
        return self.combined_branch


class ComputeMetricsBranchCombiner(DataflowPlanNodeVisitor[ComputeMetricsBranchCombinerResult]):
    """Combines branches where the leaf node is a ComputeMetricsNode.

    This considers two branches, a left branch and a right branch. The left branch is supplied via the argument in the
    initializer while the right branch is supplied via .accept(). This then attempts to create a similar branch that is
    the superposition of the two branches. For this to be possible, the two branches must be of the same structure,
    and the respective nodes in each branch must be the same type and have compatible parameters.

    For example, the superposition / combination of

    left_branch:
        <ComputeMetricsNode metrics=["bookings"]
            <AggregateMeasuresNode>
                <FilterElementsNode include_specs=["bookings"]>
                    <ReadSqlSourceNode semantic_model="bookings_source"/>
                </>
            </>
        </>
    right_branch:
        <ComputeMetricsNode metrics=["booking_value"]
            <AggregateMeasuresNode>
                <FilterElementsNode include_specs=["booking_value"]>
                    <ReadSqlSourceNode semantic_model="bookings_source"/>
                </>
            </>
        </>

    is

    <ComputeMetricsNode metrics=["bookings", "booking_value"]
        <AggregateMeasuresNode>
            <FilterElementsNode include_specs=["bookings", "booking_value"]>
                <ReadSqlSourceNode semantic_model="bookings_source"/>
            </>
        </>
    </>

    However,

    left_branch:
        <ComputeMetricsNode metrics=["bookings"]
            <AggregateMeasuresNode>
                <FilterElementsNode include_specs=["bookings", "is_instant"]>
                    <ReadSqlSourceNode semantic_model="bookings_source"/>
                </>
            </>
        </>
    right_branch:
        <ComputeMetricsNode metrics=["booking_value"]
            <AggregateMeasuresNode>
                <FilterElementsNode include_specs=["booking_value"]>
                    <ReadSqlSourceNode semantic_model="bookings_source"/>
                </>
            </>
        </>

    can't be superpositioned / combined because the different set of linkable specs in the FilterElementsNode will cause
    the AggregatedMeasuresNode to produce values for the measure that is different from the original branches. The logic
    to determine whether this is possible for each node type is encapsulated into the handler for each node type.

    In general, the questions to consider for combination are:

    * For nodes with no parents, can you combine the corresponding left and right nodes to produce an output
    that is has the same linkable specs, but a superset of the measures / metrics?
    * For other nodes, does a set of inputs that have the same linkable specs as the inputs to the left and right nodes
    but a superset of measures / metrics produce an output that superset of measures / metrics (with the same values)
    as the left and right nodes?

    The visitor traverses the dataflow plan via DFS, recursively combining the parent nodes first and combining the
    current node. During each step, current_left_node and current_right node always points to nodes that should be
    correlated between the two branches. If the structure is different or if combination can't happen due to node
    differences between the branches, a ComputeMetricsBranchCombinerResult (indicating that combination not possible)
    is propagated up to the result at the root node.
    """

    def __init__(self, left_branch_node: DataflowPlanNode) -> None:  # noqa: D107
        self._current_left_node: DataflowPlanNode = left_branch_node

    def _log_visit_node_type(self, node: DataflowPlanNode) -> None:
        logger.debug(LazyFormat(lambda: f"Visiting {node.node_id}"))

    def _log_combine_failure(
        self,
        left_node: DataflowPlanNode,
        right_node: DataflowPlanNode,
        combine_failure_reason: str,
    ) -> None:
        logger.debug(
            LazyFormat(
                "Unable to combine nodes",
                combine_failure_reason=combine_failure_reason,
                left_node=left_node.node_id,
                right_node=right_node.node_id,
            )
        )

    def _log_combine_success(
        self,
        left_node: DataflowPlanNode,
        right_node: DataflowPlanNode,
        combined_node: DataflowPlanNode,
    ) -> None:
        logger.debug(
            LazyFormat(
                "Successfully combined nodes",
                left_node=left_node.node_id,
                right_node=right_node.node_id,
                combined_node=combined_node.node_id,
            )
        )

    def _combine_parent_branches(self, current_right_node: DataflowPlanNode) -> Optional[Sequence[DataflowPlanNode]]:
        if len(self._current_left_node.parent_nodes) != len(current_right_node.parent_nodes):
            self._log_combine_failure(
                left_node=self._current_left_node,
                right_node=current_right_node,
                combine_failure_reason="parent counts are unequal",
            )
            return None

        results_of_visiting_parent_nodes: List[ComputeMetricsBranchCombinerResult] = []

        for i, right_node_parent_node in enumerate(current_right_node.parent_nodes):
            left_position_before_recursion = self._current_left_node
            self._current_left_node = self._current_left_node.parent_nodes[i]
            results_of_visiting_parent_nodes.append(right_node_parent_node.accept(self))
            self._current_left_node = left_position_before_recursion

        combined_parents: List[DataflowPlanNode] = []
        for result in results_of_visiting_parent_nodes:
            if result.combined_branch is None:
                self._log_combine_failure(
                    left_node=self._current_left_node,
                    right_node=current_right_node,
                    combine_failure_reason="not all parents could be combined",
                )
                return None
            combined_parents.append(result.combined_branch)

        return combined_parents

    def _default_handler(self, current_right_node: DataflowPlanNode) -> ComputeMetricsBranchCombinerResult:
        combined_parent_nodes = self._combine_parent_branches(current_right_node)
        if combined_parent_nodes is None:
            return ComputeMetricsBranchCombinerResult()

        new_parent_nodes = combined_parent_nodes

        # If the parent nodes were combined, and the left node is the same as the right node, then the left and right
        # nodes can be combined.
        if self._current_left_node.functionally_identical(current_right_node):
            combined_node = current_right_node.with_new_parents(new_parent_nodes)
            self._log_combine_success(
                left_node=self._current_left_node, right_node=current_right_node, combined_node=combined_node
            )
            return ComputeMetricsBranchCombinerResult(combined_node)

        self._log_combine_failure(
            left_node=self._current_left_node,
            right_node=current_right_node,
            combine_failure_reason="there are functional differences",
        )
        return ComputeMetricsBranchCombinerResult()

    def visit_source_node(self, node: ReadSqlSourceNode) -> ComputeMetricsBranchCombinerResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_join_on_entities_node(  # noqa: D102
        self, node: JoinOnEntitiesNode
    ) -> ComputeMetricsBranchCombinerResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_aggregate_measures_node(  # noqa: D102
        self, node: AggregateMeasuresNode
    ) -> ComputeMetricsBranchCombinerResult:  # noqa: D102
        self._log_visit_node_type(node)
        current_right_node = node

        combined_parent_nodes = self._combine_parent_branches(current_right_node)
        if combined_parent_nodes is None:
            return ComputeMetricsBranchCombinerResult()

        if not isinstance(self._current_left_node, current_right_node.__class__):
            self._log_combine_failure(
                left_node=self._current_left_node,
                right_node=current_right_node,
                combine_failure_reason=f"left node is not a {current_right_node.__class__.__name__}",
            )
            return ComputeMetricsBranchCombinerResult()

        assert len(combined_parent_nodes) == 1
        combined_parent_node = combined_parent_nodes[0]

        combined_metric_input_measure_specs = tuple(
            dict.fromkeys(
                self._current_left_node.metric_input_measure_specs + current_right_node.metric_input_measure_specs
            ).keys()
        )

        # Avoid combining branches if the AggregateMeasuresNode specifies a metric with an alias to avoid
        # collisions e.g. two metrics use the same alias for two different measures. This is not always the case,
        # so this could be improved later.
        seen_aliases = set()
        for spec in combined_metric_input_measure_specs:
            if spec.alias is not None:
                if spec.alias in seen_aliases:
                    self._log_combine_failure(
                        left_node=self._current_left_node,
                        right_node=current_right_node,
                        combine_failure_reason=f"Found multiple metric input measure specs with alias '{spec.alias}'",
                    )
                    return ComputeMetricsBranchCombinerResult()
                seen_aliases.add(spec.alias)

        combined_node = AggregateMeasuresNode.create(
            parent_node=combined_parent_node,
            metric_input_measure_specs=combined_metric_input_measure_specs,
        )
        self._log_combine_success(
            left_node=self._current_left_node,
            right_node=current_right_node,
            combined_node=combined_node,
        )
        return ComputeMetricsBranchCombinerResult(combined_node)

    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> ComputeMetricsBranchCombinerResult:  # noqa: D102
        current_right_node = node
        self._log_visit_node_type(current_right_node)
        combined_parent_nodes = self._combine_parent_branches(current_right_node)
        if combined_parent_nodes is None:
            return ComputeMetricsBranchCombinerResult()

        if not isinstance(self._current_left_node, current_right_node.__class__):
            self._log_combine_failure(
                left_node=self._current_left_node,
                right_node=current_right_node,
                combine_failure_reason=f"left node is not a {current_right_node.__class__.__name__}",
            )
            return ComputeMetricsBranchCombinerResult()

        can_combine, combine_failure_reason = self._current_left_node.can_combine(current_right_node)
        if not can_combine:
            self._log_combine_failure(
                left_node=self._current_left_node,
                right_node=current_right_node,
                combine_failure_reason=combine_failure_reason,
            )
            return ComputeMetricsBranchCombinerResult()

        assert len(combined_parent_nodes) == 1
        combined_parent_node = combined_parent_nodes[0]
        assert combined_parent_node is not None

        # Dedupe (preserving order for output consistency) as it's possible for multiple derived metrics to use the same
        # metric.
        unique_metric_specs: List[MetricSpec] = []
        for metric_spec in tuple(self._current_left_node.metric_specs) + tuple(current_right_node.metric_specs):
            if metric_spec not in unique_metric_specs:
                unique_metric_specs.append(metric_spec)

        combined_node = ComputeMetricsNode.create(
            parent_node=combined_parent_node,
            metric_specs=unique_metric_specs,
            aggregated_to_elements=current_right_node.aggregated_to_elements,
            for_group_by_source_node=current_right_node.for_group_by_source_node,
        )
        self._log_combine_success(
            left_node=self._current_left_node,
            right_node=current_right_node,
            combined_node=combined_node,
        )
        return ComputeMetricsBranchCombinerResult(combined_node)

    def _handle_unsupported_node(self, current_right_node: DataflowPlanNode) -> ComputeMetricsBranchCombinerResult:
        self._log_combine_failure(
            left_node=self._current_left_node,
            right_node=current_right_node,
            combine_failure_reason=(
                f"right node is of type {current_right_node.__class__.__name__} which is not yet handled"
            ),
        )
        return ComputeMetricsBranchCombinerResult()

    def visit_window_reaggregation_node(  # noqa: D102
        self, node: WindowReaggregationNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._handle_unsupported_node(node)

    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> ComputeMetricsBranchCombinerResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._handle_unsupported_node(node)

    def visit_where_constraint_node(  # noqa: D102
        self, node: WhereConstraintNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_write_to_result_data_table_node(  # noqa: D102
        self, node: WriteToResultDataTableNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._handle_unsupported_node(node)

    def visit_write_to_result_table_node(  # noqa: D102
        self, node: WriteToResultTableNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._handle_unsupported_node(node)

    def visit_filter_elements_node(self, node: FilterElementsNode) -> ComputeMetricsBranchCombinerResult:  # noqa: D102
        self._log_visit_node_type(node)

        current_right_node = node
        results_of_visiting_parent_nodes = self._combine_parent_branches(current_right_node)
        if results_of_visiting_parent_nodes is None:
            return ComputeMetricsBranchCombinerResult()

        if not isinstance(self._current_left_node, current_right_node.__class__):
            self._log_combine_failure(
                left_node=self._current_left_node,
                right_node=current_right_node,
                combine_failure_reason=f"left node is not a {current_right_node.__class__.__name__}",
            )
            return ComputeMetricsBranchCombinerResult()

        assert len(results_of_visiting_parent_nodes) == 1
        combined_parent_node = results_of_visiting_parent_nodes[0]
        assert combined_parent_node is not None

        # For the FilterElementsNode to be combined, the linkable specs have to be the same for the left and right.
        if not MatchingLinkableSpecsTransform(self._current_left_node.include_specs).transform(
            current_right_node.include_specs
        ):
            self._log_combine_failure(
                left_node=self._current_left_node,
                right_node=current_right_node,
                combine_failure_reason="linkable specs in the filter do not match",
            )
            return ComputeMetricsBranchCombinerResult()

        # De-dupe so that we don't see the same spec twice in include specs. For example, this can happen with dimension
        # specs since any branch that is merged together needs to output the same set of dimensions.
        combined_node = FilterElementsNode.create(
            parent_node=combined_parent_node,
            include_specs=self._current_left_node.include_specs.merge(current_right_node.include_specs).dedupe(),
        )
        self._log_combine_success(
            left_node=self._current_left_node,
            right_node=current_right_node,
            combined_node=combined_node,
        )
        return ComputeMetricsBranchCombinerResult(combined_node)

    def visit_combine_aggregated_outputs_node(  # noqa: D102
        self, node: CombineAggregatedOutputsNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._handle_unsupported_node(node)

    def visit_constrain_time_range_node(  # noqa: D102
        self, node: ConstrainTimeRangeNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_join_over_time_range_node(  # noqa: D102
        self, node: JoinOverTimeRangeNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_semi_additive_join_node(  # noqa: D102
        self, node: SemiAdditiveJoinNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_metric_time_dimension_transform_node(  # noqa: D102
        self, node: MetricTimeDimensionTransformNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_join_to_time_spine_node(  # noqa: D102
        self, node: JoinToTimeSpineNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_add_generated_uuid_column_node(  # noqa: D102
        self, node: AddGeneratedUuidColumnNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_join_conversion_events_node(  # noqa: D102
        self, node: JoinConversionEventsNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_join_to_custom_granularity_node(  # noqa: D102
        self, node: JoinToCustomGranularityNode
    ) -> ComputeMetricsBranchCombinerResult:
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_min_max_node(self, node: MinMaxNode) -> ComputeMetricsBranchCombinerResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_alias_specs_node(self, node: AliasSpecsNode) -> ComputeMetricsBranchCombinerResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)



from __future__ import annotations

from typing import Dict, Iterable, Tuple, TypeVar

IterableT = TypeVar("IterableT")


def ordered_dedupe(*iterables: Iterable[IterableT]) -> Tuple[IterableT, ...]:
    """De-duplicates the items in the iterables while preserving the order."""
    ordered_results: Dict[IterableT, None] = {}
    for iterable in iterables:
        for item in iterable:
            ordered_results[item] = None

    return tuple(ordered_results.keys())



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Set, Tuple

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.visitor import VisitorOutputT
from typing_extensions import override

from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class ComputeMetricsNode(DataflowPlanNode):
    """A node that computes metrics from input measures. Dimensions / entities are passed through.

    Attributes:
        metric_specs: The specs for the metrics that this should compute.
        for_group_by_source_node: Whether the node is part of a dataflow plan used for a group by source node.
    """

    metric_specs: Tuple[MetricSpec, ...]
    for_group_by_source_node: bool
    _aggregated_to_elements: Tuple[LinkableInstanceSpec, ...]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()

        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        metric_specs: Sequence[MetricSpec],
        aggregated_to_elements: Set[LinkableInstanceSpec],
        for_group_by_source_node: bool = False,
    ) -> ComputeMetricsNode:
        return ComputeMetricsNode(
            parent_nodes=(parent_node,),
            metric_specs=tuple(metric_specs),
            _aggregated_to_elements=tuple(aggregated_to_elements),
            for_group_by_source_node=for_group_by_source_node,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_COMPUTE_METRICS_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_compute_metrics_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Compute Metrics via Expressions"""

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        displayed_properties = tuple(super().displayed_properties) + tuple(
            DisplayedProperty("metric_spec", metric_spec) for metric_spec in self.metric_specs
        )
        if self.for_group_by_source_node:
            displayed_properties += (DisplayedProperty("for_group_by_source_node", self.for_group_by_source_node),)
        return displayed_properties

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        if not isinstance(other_node, self.__class__):
            return False

        if other_node.metric_specs != self.metric_specs:
            return False

        return (
            isinstance(other_node, self.__class__)
            and other_node.metric_specs == self.metric_specs
            and other_node.aggregated_to_elements == self.aggregated_to_elements
            and other_node.for_group_by_source_node == self.for_group_by_source_node
        )

    def can_combine(self, other_node: ComputeMetricsNode) -> Tuple[bool, str]:
        """Check certain node attributes against another node to determine if the two can be combined.

        Return a bool and a string reason for the failure to combine (if applicable) to be used in logging for
        ComputeMetricsBranchCombiner.
        """
        if not other_node.aggregated_to_elements == self.aggregated_to_elements:
            return False, "nodes are aggregated to different elements"

        if other_node.for_group_by_source_node != self.for_group_by_source_node:
            return False, "one node is a group by metric source node"

        alias_to_metric_spec = {spec.alias: spec for spec in self.metric_specs if spec.alias is not None}

        for spec in other_node.metric_specs:
            if (
                spec.alias is not None
                and spec.alias in alias_to_metric_spec
                and alias_to_metric_spec[spec.alias] != spec
            ):
                return (
                    False,
                    f"Alias '{spec.alias}' is defined in both nodes but it refers to different things in each of them",
                )

        return True, ""

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> ComputeMetricsNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return ComputeMetricsNode.create(
            parent_node=new_parent_nodes[0],
            metric_specs=self.metric_specs,
            for_group_by_source_node=self.for_group_by_source_node,
            aggregated_to_elements=self.aggregated_to_elements,
        )

    @property
    @override
    def aggregated_to_elements(self) -> Set[LinkableInstanceSpec]:
        return set(self._aggregated_to_elements)



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass

from metricflow_semantics.specs.where_filter.where_filter_spec import WhereFilterSpec


@dataclass(frozen=True)
class WhereFilterSpecSet(SerializableDataclass):
    """Class to encapsulate filters needed at a certain point of the queried metric.

    This class splits up the filters based on where it was defined at, which can then be used to
    determine where each filter should be applied during each step of the metric building process.

    measure-level: filters defined on the input_measure
    metric-level: filters defined on the input_metric or metric:filter
    query-level: filters defined at query time
    """

    measure_level_filter_specs: Tuple[WhereFilterSpec, ...] = ()
    metric_level_filter_specs: Tuple[WhereFilterSpec, ...] = ()
    query_level_filter_specs: Tuple[WhereFilterSpec, ...] = ()

    @property
    def after_measure_aggregation_filter_specs(self) -> Tuple[WhereFilterSpec, ...]:
        """Returns filters relevant to post-measure aggregation."""
        return self.metric_level_filter_specs + self.query_level_filter_specs

    @property
    def all_filter_specs(self) -> Tuple[WhereFilterSpec, ...]:
        """Returns all the filters in this class.

        Generally, before measure aggregation, all filters should be applied.
        """
        return self.measure_level_filter_specs + self.metric_level_filter_specs + self.query_level_filter_specs

    def merge(self, other: WhereFilterSpecSet) -> WhereFilterSpecSet:
        """Merge 2 WhereFilterSpecSet together."""
        return WhereFilterSpecSet(
            measure_level_filter_specs=self.measure_level_filter_specs + other.measure_level_filter_specs,
            metric_level_filter_specs=self.metric_level_filter_specs + other.metric_level_filter_specs,
            query_level_filter_specs=self.query_level_filter_specs + other.query_level_filter_specs,
        )



from __future__ import annotations

import logging
from typing import Dict

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

from metricflow.sql.sql_plan import SqlCteAliasMapping, SqlSelectStatementNode

logger = logging.getLogger(__name__)


class SqlCteAliasMappingLookup:
    """A mutable lookup that stores the CTE-alias mapping at a given node.

    In cases with nested CTEs in a SELECT, it's possible that a CTE defined in an inner SELECT has an alias that is the
    same as a CTE defined in an outer SELECT. e.g.

        # outer_cte
        WITH cte_0 AS (
            SELECT 1 AS col_0
        )

        # outer_select
        SELECT col_0
        FROM (

            # inner_cte
            WITH cte_0 AS (
                SELECT 2 AS col_0
            )
            # inner_select
            SELECT col_0 FROM cte_0
        )
        ...

    In this case, `outer_cte` and `inner_cte` both have the same alias `cte_0`. When `cte_0` is referenced from
    `inner_select`, it is referring to the `inner_cte`. For column pruning, it is necessary to figure out which CTE
    a given alias is referencing, so this class helps to keep track of that mapping.
    """

    def __init__(self) -> None:  # noqa: D107
        self._select_node_to_cte_alias_mapping: Dict[SqlSelectStatementNode, SqlCteAliasMapping] = {}

    def cte_alias_mapping_exists(self, select_node: SqlSelectStatementNode) -> bool:
        """Returns true if the CTE-alias mapping for the given node has been recorded."""
        return select_node in self._select_node_to_cte_alias_mapping

    def add_cte_alias_mapping(
        self,
        select_node: SqlSelectStatementNode,
        cte_alias_mapping: SqlCteAliasMapping,
    ) -> None:
        """Associate the given CTE-alias mapping with the given node.

        Raises an exception if a mapping already exists.
        """
        if select_node in self._select_node_to_cte_alias_mapping:
            raise RuntimeError(
                str(
                    LazyFormat(
                        "`select_node` node has already been added,",
                        # child_select_node=child_select_node,
                        select_node=select_node,
                        current_mapping=self._select_node_to_cte_alias_mapping,
                    )
                )
            )

        self._select_node_to_cte_alias_mapping[select_node] = cte_alias_mapping

    def get_cte_alias_mapping(self, select_node: SqlSelectStatementNode) -> SqlCteAliasMapping:
        """Return the CTE-alias mapping for the given node.

        Raises an exception if a mapping was not previously added for the given node.
        """
        cte_alias_mapping = self._select_node_to_cte_alias_mapping.get(select_node)
        if cte_alias_mapping is None:
            raise RuntimeError(
                str(LazyFormat("CTE alias mapping does not exist for the given `select_node`", select_node=select_node))
            )
        return cte_alias_mapping



from __future__ import annotations

from dataclasses import dataclass

from metricflow.execution.execution_plan import ExecutionPlan
from metricflow.plan_conversion.convert_to_sql_plan import ConvertToSqlPlanResult
from metricflow.sql.render.sql_plan_renderer import SqlPlanRenderResult


@dataclass(frozen=True)
class ConvertToExecutionPlanResult:
    """A result object for returning the results of converting a dataflow plan into an execution plan."""

    convert_to_sql_plan_result: ConvertToSqlPlanResult
    render_sql_result: SqlPlanRenderResult
    execution_plan: ExecutionPlan



from __future__ import annotations

import logging

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from typing_extensions import override

from metricflow.dataflow.dataflow_plan import (
    DataflowPlan,
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor
from metricflow.dataflow.nodes.add_generated_uuid import AddGeneratedUuidColumnNode
from metricflow.dataflow.nodes.aggregate_measures import AggregateMeasuresNode
from metricflow.dataflow.nodes.alias_specs import AliasSpecsNode
from metricflow.dataflow.nodes.combine_aggregated_outputs import CombineAggregatedOutputsNode
from metricflow.dataflow.nodes.compute_metrics import ComputeMetricsNode
from metricflow.dataflow.nodes.constrain_time import ConstrainTimeRangeNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataflow.nodes.join_conversion_events import JoinConversionEventsNode
from metricflow.dataflow.nodes.join_over_time import JoinOverTimeRangeNode
from metricflow.dataflow.nodes.join_to_base import JoinOnEntitiesNode
from metricflow.dataflow.nodes.join_to_custom_granularity import JoinToCustomGranularityNode
from metricflow.dataflow.nodes.join_to_time_spine import JoinToTimeSpineNode
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataflow.nodes.min_max import MinMaxNode
from metricflow.dataflow.nodes.order_by_limit import OrderByLimitNode
from metricflow.dataflow.nodes.read_sql_source import ReadSqlSourceNode
from metricflow.dataflow.nodes.semi_additive_join import SemiAdditiveJoinNode
from metricflow.dataflow.nodes.where_filter import WhereConstraintNode
from metricflow.dataflow.nodes.window_reaggregation_node import WindowReaggregationNode
from metricflow.dataflow.nodes.write_to_data_table import WriteToResultDataTableNode
from metricflow.dataflow.nodes.write_to_table import WriteToResultTableNode
from metricflow.execution.convert_to_execution_plan import ConvertToExecutionPlanResult
from metricflow.execution.execution_plan import (
    ExecutionPlan,
    SelectSqlQueryToDataTableTask,
    SelectSqlQueryToTableTask,
    SqlStatement,
)
from metricflow.plan_conversion.convert_to_sql_plan import ConvertToSqlPlanResult
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlPlanConverter
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.optimizer.optimization_levels import SqlOptimizationLevel
from metricflow.sql.render.sql_plan_renderer import SqlPlanRenderer, SqlPlanRenderResult

logger = logging.getLogger(__name__)


class DataflowToExecutionPlanConverter(DataflowPlanNodeVisitor[ConvertToExecutionPlanResult]):
    """Converts a dataflow plan to an execution plan."""

    def __init__(
        self,
        sql_plan_converter: DataflowToSqlPlanConverter,
        sql_plan_renderer: SqlPlanRenderer,
        sql_client: SqlClient,
        sql_optimization_level: SqlOptimizationLevel,
    ) -> None:
        """Constructor.

        Args:
            sql_plan_converter: Converts a dataflow plan node to a SQL query plan
            sql_plan_renderer: Converts a SQL query plan to SQL text
            sql_client: The client to use for running queries.
            sql_optimization_level: The optimization level to use for generating the SQL.
        """
        self._sql_plan_converter = sql_plan_converter
        self._sql_plan_renderer = sql_plan_renderer
        self._sql_client = sql_client
        self._optimization_level = sql_optimization_level

    def _convert_to_sql_plan(self, node: DataflowPlanNode) -> ConvertToSqlPlanResult:
        logger.debug(LazyFormat(lambda: f"Generating SQL query plan from {node.node_id}"))
        result = self._sql_plan_converter.convert_to_sql_plan(
            sql_engine_type=self._sql_client.sql_engine_type,
            optimization_level=self._optimization_level,
            dataflow_plan_node=node,
        )
        logger.debug(LazyFormat(lambda: f"Generated SQL query plan is:\n{result.sql_plan.structure_text()}"))
        return result

    def _render_sql(self, convert_to_sql_plan_result: ConvertToSqlPlanResult) -> SqlPlanRenderResult:
        return self._sql_plan_renderer.render_sql_plan(convert_to_sql_plan_result.sql_plan)

    @override
    def visit_write_to_result_data_table_node(self, node: WriteToResultDataTableNode) -> ConvertToExecutionPlanResult:
        convert_to_sql_plan_result = self._convert_to_sql_plan(node)
        render_sql_result = self._render_sql(convert_to_sql_plan_result)
        execution_plan = ExecutionPlan(
            leaf_tasks=(
                SelectSqlQueryToDataTableTask.create(
                    sql_client=self._sql_client,
                    sql_statement=SqlStatement(render_sql_result.sql, render_sql_result.bind_parameter_set),
                ),
            )
        )
        return ConvertToExecutionPlanResult(
            convert_to_sql_plan_result=convert_to_sql_plan_result,
            render_sql_result=render_sql_result,
            execution_plan=execution_plan,
        )

    @override
    def visit_write_to_result_table_node(self, node: WriteToResultTableNode) -> ConvertToExecutionPlanResult:
        convert_to_sql_plan_result = self._convert_to_sql_plan(node)
        render_sql_result = self._render_sql(convert_to_sql_plan_result)
        execution_plan = ExecutionPlan(
            leaf_tasks=(
                SelectSqlQueryToTableTask.create(
                    sql_client=self._sql_client,
                    sql_statement=SqlStatement(
                        sql=render_sql_result.sql,
                        bind_parameter_set=render_sql_result.bind_parameter_set,
                    ),
                    output_table=node.output_sql_table,
                ),
            ),
        )
        return ConvertToExecutionPlanResult(
            convert_to_sql_plan_result=convert_to_sql_plan_result,
            render_sql_result=render_sql_result,
            execution_plan=execution_plan,
        )

    def convert_to_execution_plan(self, dataflow_plan: DataflowPlan) -> ConvertToExecutionPlanResult:
        """Convert the dataflow plan to an execution plan."""
        assert len(dataflow_plan.sink_nodes) == 1, "Only 1 sink node in the plan is currently supported."
        return dataflow_plan.sink_nodes[0].accept(self)

    @override
    def visit_source_node(self, node: ReadSqlSourceNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_join_on_entities_node(self, node: JoinOnEntitiesNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_aggregate_measures_node(self, node: AggregateMeasuresNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_window_reaggregation_node(self, node: WindowReaggregationNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_where_constraint_node(self, node: WhereConstraintNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_filter_elements_node(self, node: FilterElementsNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_combine_aggregated_outputs_node(self, node: CombineAggregatedOutputsNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_semi_additive_join_node(self, node: SemiAdditiveJoinNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_metric_time_dimension_transform_node(
        self, node: MetricTimeDimensionTransformNode
    ) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_join_to_time_spine_node(self, node: JoinToTimeSpineNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_min_max_node(self, node: MinMaxNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_add_generated_uuid_column_node(self, node: AddGeneratedUuidColumnNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_join_conversion_events_node(self, node: JoinConversionEventsNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_join_to_custom_granularity_node(self, node: JoinToCustomGranularityNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError

    @override
    def visit_alias_specs_node(self, node: AliasSpecsNode) -> ConvertToExecutionPlanResult:
        raise NotImplementedError



from __future__ import annotations

import threading
from typing import TYPE_CHECKING, Callable, Dict, Generic, Optional, TypeVar

if TYPE_CHECKING:
    # Hack: ensure type checking is not erased for parameters in methods decorated with @lru_cache.
    F = TypeVar("F", bound=Callable)

    def typed_lru_cache(f: F) -> F:  # noqa: D103
        pass

else:
    from functools import lru_cache as typed_lru_cache  # noqa: F401

KeyT = TypeVar("KeyT")
ValueT = TypeVar("ValueT")


class LruCache(Generic[KeyT, ValueT]):
    """An LRU cache based on the insertion order of dictionaries.

    Since Python dictionaries iterate in the order that keys were inserted, they are used as the basis of this cache.
    When an item is retrieved, the item in the dictionary is removed then re-inserted.

    This cache is used instead of the `fuctools.lru_cache` decorator for class instance methods as `lru_cache` keeps a
    reference to the instance, preventing garbage collection of the instance using the decorator until the eviction of
    the associated entry.
    """

    def __init__(self, max_cache_items: int, cache_dict: Optional[Dict[KeyT, ValueT]] = None) -> None:
        """Initializer.

        Args:
            max_cache_items: Limit of cache items to store. Once the limit is hit, the oldest item is evicted.
            cache_dict: For shared use cases - the dictionary to use for the cache.
        """
        self._lock = threading.Lock()
        self._max_cache_items = max_cache_items
        self._cache_dict: Dict[KeyT, ValueT] = cache_dict or {}

    def get(self, key: KeyT) -> Optional[ValueT]:  # noqa: D102
        with self._lock:
            value = self._cache_dict.get(key)
            if value is not None:
                del self._cache_dict[key]
                self._cache_dict[key] = value
                return value

            return None

    def set(self, key: KeyT, value: ValueT) -> None:  # noqa: D102
        with self._lock:
            if key in self._cache_dict:
                return

            while len(self._cache_dict) >= self._max_cache_items:
                key_to_delete = next(iter(self._cache_dict))
                del self._cache_dict[key_to_delete]

            self._cache_dict[key] = value

    def copy(self) -> LruCache:  # noqa: D102
        return LruCache(max_cache_items=self._max_cache_items, cache_dict=dict(self._cache_dict))



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Tuple

from dbt_semantic_interfaces.naming.keywords import METRIC_TIME_ELEMENT_NAME
from dbt_semantic_interfaces.protocols import MetricInput
from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class OffsetMetricRequiresMetricTimeIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue where metric_time is required, but not included in the query.

    metric_time is required as a group-by-item when querying offset metrics.
    """

    metric_reference: MetricReference
    input_metrics: Tuple[MetricInput, ...]

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return (
            f"The query includes a metric {repr(self.metric_reference.element_name)} that specifies a time offset in "
            f"input metrics: {repr(self.input_metrics)}. However, group-by-items do not include "
            f"{repr(METRIC_TIME_ELEMENT_NAME)}."
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> OffsetMetricRequiresMetricTimeIssue:
        return OffsetMetricRequiresMetricTimeIssue(
            issue_type=self.issue_type,
            parent_issues=self.parent_issues,
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            metric_reference=self.metric_reference,
            input_metrics=self.input_metrics,
        )

    @staticmethod
    def from_parameters(  # noqa: D102
        metric_reference: MetricReference,
        input_metrics: Sequence[MetricInput],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> OffsetMetricRequiresMetricTimeIssue:
        return OffsetMetricRequiresMetricTimeIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            query_resolution_path=query_resolution_path,
            metric_reference=metric_reference,
            input_metrics=tuple(input_metrics),
        )



from __future__ import annotations

import datetime
from typing import Optional

import dateutil.relativedelta
from dateutil.relativedelta import relativedelta
from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.type_enums import TimeGranularity
from typing_extensions import override

from metricflow_semantics.filters.time_constraint import TimeRangeConstraint
from metricflow_semantics.time.time_period import TimePeriodAdjuster


class DateutilTimePeriodAdjuster(TimePeriodAdjuster):
    """Implementation of time period adjustments using `dateutil`.

    * `relativedelta` will not change weekday if already at the given weekday, even with a Nth parameter.
    * `relativedelta` will automatically handle day values that exceed the number of days in months with < 31 days.
    """

    def _relative_delta_for_window(self, time_granularity: TimeGranularity, count: int) -> relativedelta:
        """Relative-delta to cover time windows specified at different grains."""
        if (
            time_granularity is TimeGranularity.NANOSECOND
            or time_granularity is TimeGranularity.MICROSECOND
            or time_granularity is TimeGranularity.MILLISECOND
        ):
            raise ValueError("Time constraints only support SECOND or larger time granularities.")
        elif time_granularity is TimeGranularity.SECOND:
            return relativedelta(seconds=count)
        elif time_granularity is TimeGranularity.MINUTE:
            return relativedelta(minutes=count)
        elif time_granularity is TimeGranularity.HOUR:
            return relativedelta(hours=count)
        elif time_granularity is TimeGranularity.DAY:
            return relativedelta(days=count)
        elif time_granularity is TimeGranularity.WEEK:
            return relativedelta(weeks=count)
        elif time_granularity is TimeGranularity.MONTH:
            return relativedelta(months=count)
        elif time_granularity is TimeGranularity.QUARTER:
            return relativedelta(months=count * 3)
        elif time_granularity is TimeGranularity.YEAR:
            return relativedelta(years=count)
        else:
            assert_values_exhausted(time_granularity)

    @override
    def expand_time_constraint_to_fill_granularity(
        self, time_constraint: TimeRangeConstraint, granularity: TimeGranularity
    ) -> TimeRangeConstraint:
        adjusted_start = self.adjust_to_start_of_period(granularity, time_constraint.start_time)
        adjusted_end = self.adjust_to_end_of_period(granularity, time_constraint.end_time)

        if adjusted_start < TimeRangeConstraint.ALL_TIME_BEGIN():
            adjusted_start = TimeRangeConstraint.ALL_TIME_BEGIN()
        if adjusted_end > TimeRangeConstraint.ALL_TIME_END():
            adjusted_end = TimeRangeConstraint.ALL_TIME_END()

        return TimeRangeConstraint(start_time=adjusted_start, end_time=adjusted_end)

    @override
    def adjust_to_start_of_period(
        self, time_granularity: TimeGranularity, date_to_adjust: datetime.datetime
    ) -> datetime.datetime:
        """Takes a date and adjusts it to the start of the given granularity period.

        Used to adjust time constraints when the time dimension they will be applied to has a larger granularity
        than the time constraint requested.

        Args:
            time_granularity: The granularity of the time dimension that will be constrained.
            date_to_adjust: The time constraint start date.
        """
        # Strip time data that is irrelevant to avoid rendering an unnecessarily complex time constraint.
        date_to_adjust = date_to_adjust.replace(microsecond=0)
        if time_granularity.to_int() >= TimeGranularity.DAY.to_int():
            date_to_adjust = date_to_adjust.replace(hour=0, minute=0, second=0)

        if (
            time_granularity is TimeGranularity.NANOSECOND
            or time_granularity is TimeGranularity.MICROSECOND
            or time_granularity is TimeGranularity.MILLISECOND
        ):
            raise ValueError("Time constraints only support SECOND and larger time granularities.")
        elif time_granularity is TimeGranularity.SECOND:
            return date_to_adjust
        elif time_granularity is TimeGranularity.MINUTE:
            return date_to_adjust + relativedelta(second=0)
        elif time_granularity is TimeGranularity.HOUR:
            return date_to_adjust + relativedelta(minute=0, second=0)
        elif time_granularity is TimeGranularity.DAY:
            return date_to_adjust
        elif time_granularity is TimeGranularity.WEEK:
            return date_to_adjust + relativedelta(weekday=dateutil.relativedelta.MO(-1))
        elif time_granularity is TimeGranularity.MONTH:
            return date_to_adjust + relativedelta(day=1)
        elif time_granularity is TimeGranularity.QUARTER:
            if date_to_adjust.month <= 3:
                return date_to_adjust + relativedelta(month=1, day=1)
            elif date_to_adjust.month <= 6:
                return date_to_adjust + relativedelta(month=4, day=1)
            elif date_to_adjust.month <= 9:
                return date_to_adjust + relativedelta(month=7, day=1)
            else:
                return date_to_adjust + relativedelta(month=10, day=1)
        elif time_granularity is TimeGranularity.YEAR:
            return date_to_adjust + relativedelta(month=1, day=1)
        else:
            assert_values_exhausted(time_granularity)

    @override
    def adjust_to_end_of_period(
        self, time_granularity: TimeGranularity, date_to_adjust: datetime.datetime
    ) -> datetime.datetime:
        """Takes a date and adjusts it to the end of the given granularity period.

        Used to adjust time constraints when the time dimension they will be applied to has a larger granularity
        than the time constraint requested.

        Args:
            time_granularity: The granularity of the time dimension that will be constrained.
            date_to_adjust: The time constraint end date.
        """
        # Strip time data that is irrelevant to avoid rendering an unnecessarily complex time constraint.
        date_to_adjust = date_to_adjust.replace(microsecond=0)
        if time_granularity.to_int() >= TimeGranularity.DAY.to_int():
            date_to_adjust = date_to_adjust.replace(hour=0, minute=0, second=0)

        if (
            time_granularity is TimeGranularity.NANOSECOND
            or time_granularity is TimeGranularity.MICROSECOND
            or time_granularity is TimeGranularity.MILLISECOND
        ):
            raise ValueError("Time constraints only support SECOND and larger time granularities.")
        elif time_granularity is TimeGranularity.SECOND:
            return date_to_adjust
        elif time_granularity is TimeGranularity.MINUTE:
            return date_to_adjust + relativedelta(second=59)
        elif time_granularity is TimeGranularity.HOUR:
            return date_to_adjust + relativedelta(minute=59, second=59)
        elif time_granularity is TimeGranularity.DAY:
            return date_to_adjust
        elif time_granularity is TimeGranularity.WEEK:
            return date_to_adjust + relativedelta(weekday=dateutil.relativedelta.SU(1))
        elif time_granularity is TimeGranularity.MONTH:
            return date_to_adjust + relativedelta(day=31)
        elif time_granularity is TimeGranularity.QUARTER:
            if date_to_adjust.month <= 3:
                return date_to_adjust + relativedelta(month=3, day=31)
            elif date_to_adjust.month <= 6:
                return date_to_adjust + relativedelta(month=6, day=31)
            elif date_to_adjust.month <= 9:
                return date_to_adjust + relativedelta(month=9, day=31)
            else:
                return date_to_adjust + relativedelta(month=12, day=31)
        elif time_granularity is TimeGranularity.YEAR:
            return date_to_adjust + relativedelta(month=12, day=31)
        else:
            assert_values_exhausted(time_granularity)

    @override
    def expand_time_constraint_for_cumulative_metric(
        self, time_constraint: TimeRangeConstraint, granularity: Optional[TimeGranularity], count: int
    ) -> TimeRangeConstraint:
        if granularity is not None:
            return TimeRangeConstraint(
                start_time=time_constraint.start_time - self._relative_delta_for_window(granularity, count),
                end_time=time_constraint.end_time,
            )

        # if no window is specified we want to accumulate from the beginning of time
        return TimeRangeConstraint(
            start_time=TimeRangeConstraint.ALL_TIME_BEGIN(),
            end_time=time_constraint.end_time,
        )



from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import Enum, EnumMeta

from typing_extensions import override


class IdPrefix(ABC):
    """A prefix for generating IDs. e.g. prefix=foo -> id=foo_0."""

    @property
    @abstractmethod
    def str_value(self) -> str:
        """Return the string value of this ID prefix."""
        raise NotImplementedError


class EnumMetaClassHelper(ABCMeta, EnumMeta):
    """Metaclass to allow subclassing of IdPrefix / Enum.

    Without this, you'll get an error:

        Metaclass conflict: the metaclass of a derived class must be a (non-strict) subclass of the metaclasses of all
        its bases

    since Enum has a special metaclass.
    """

    pass


class StaticIdPrefix(IdPrefix, Enum, metaclass=EnumMetaClassHelper):
    """Enumerates the prefixes used for generating IDs."""

    DATAFLOW_NODE_AGGREGATE_MEASURES_ID_PREFIX = "am"
    DATAFLOW_NODE_COMPUTE_METRICS_ID_PREFIX = "cm"
    DATAFLOW_NODE_JOIN_AGGREGATED_MEASURES_BY_GROUPBY_COLUMNS_PREFIX = "jamgc"
    DATAFLOW_NODE_JOIN_TO_STANDARD_OUTPUT_ID_PREFIX = "jso"
    DATAFLOW_NODE_JOIN_SELF_OVER_TIME_RANGE_ID_PREFIX = "jotr"
    DATAFLOW_NODE_ORDER_BY_LIMIT_ID_PREFIX = "obl"
    DATAFLOW_NODE_PASS_FILTER_ELEMENTS_ID_PREFIX = "pfe"
    DATAFLOW_NODE_READ_SQL_SOURCE_ID_PREFIX = "rss"
    DATAFLOW_NODE_WHERE_CONSTRAINT_ID_PREFIX = "wcc"
    DATAFLOW_NODE_WRITE_TO_RESULT_DATA_TABLE_ID_PREFIX = "wrd"
    DATAFLOW_NODE_WRITE_TO_RESULT_TABLE_ID_PREFIX = "wrt"
    DATAFLOW_NODE_COMBINE_AGGREGATED_OUTPUTS_ID_PREFIX = "cao"
    DATAFLOW_NODE_CONSTRAIN_TIME_RANGE_ID_PREFIX = "ctr"
    DATAFLOW_NODE_SET_MEASURE_AGGREGATION_TIME = "sma"
    DATAFLOW_NODE_SEMI_ADDITIVE_JOIN_ID_PREFIX = "saj"
    DATAFLOW_NODE_JOIN_TO_TIME_SPINE_ID_PREFIX = "jts"
    DATAFLOW_NODE_JOIN_TO_CUSTOM_GRANULARITY_ID_PREFIX = "jcg"
    DATAFLOW_NODE_MIN_MAX_ID_PREFIX = "mm"
    DATAFLOW_NODE_ADD_UUID_COLUMN_PREFIX = "auid"
    DATAFLOW_NODE_JOIN_CONVERSION_EVENTS_PREFIX = "jce"
    DATAFLOW_NODE_WINDOW_REAGGREGATION_ID_PREFIX = "wr"
    DATAFLOW_NODE_ALIAS_SPECS_ID_PREFIX = "as"

    SQL_EXPR_COLUMN_REFERENCE_ID_PREFIX = "cr"
    SQL_EXPR_COMPARISON_ID_PREFIX = "cmp"
    SQL_EXPR_FUNCTION_ID_PREFIX = "fnc"
    SQL_EXPR_PERCENTILE_ID_PREFIX = "perc"
    SQL_EXPR_STRING_ID_PREFIX = "str"
    SQL_EXPR_NULL_PREFIX = "null"
    SQL_EXPR_LOGICAL_OPERATOR_PREFIX = "lo"
    SQL_EXPR_STRING_LITERAL_PREFIX = "sl"
    SQL_EXPR_IS_NULL_PREFIX = "isn"
    SQL_EXPR_CAST_TO_TIMESTAMP_PREFIX = "ctt"
    SQL_EXPR_DATE_TRUNC = "dt"
    SQL_EXPR_SUBTRACT_TIME_INTERVAL_PREFIX = "sti"
    SQL_EXPR_ADD_TIME_PREFIX = "ati"
    SQL_EXPR_EXTRACT = "ex"
    SQL_EXPR_RATIO_COMPUTATION = "rc"
    SQL_EXPR_BETWEEN_PREFIX = "betw"
    SQL_EXPR_WINDOW_FUNCTION_ID_PREFIX = "wfnc"
    SQL_EXPR_GENERATE_UUID_PREFIX = "uuid"
    SQL_EXPR_CASE_PREFIX = "case"
    SQL_EXPR_ARITHMETIC_PREFIX = "arit"
    SQL_EXPR_INTEGER_PREFIX = "int"

    SQL_PLAN_SELECT_STATEMENT_ID_PREFIX = "ss"
    SQL_PLAN_TABLE_FROM_CLAUSE_ID_PREFIX = "tfc"
    SQL_PLAN_QUERY_FROM_CLAUSE_ID_PREFIX = "qfc"
    SQL_PLAN_CREATE_TABLE_AS_ID_PREFIX = "cta"
    SQL_PLAN_COMMON_TABLE_EXPRESSION_ID_PREFIX = "cte"

    EXEC_NODE_READ_SQL_QUERY = "rsq"
    EXEC_NODE_NOOP = "noop"
    EXEC_NODE_WRITE_TO_TABLE = "wtt"

    # Group by item resolution
    GROUP_BY_ITEM_RESOLUTION_DAG = "gbir"
    QUERY_GROUP_BY_ITEM_RESOLUTION_NODE = "qr"
    METRIC_GROUP_BY_ITEM_RESOLUTION_NODE = "mtr"
    MEASURE_GROUP_BY_ITEM_RESOLUTION_NODE = "msr"
    VALUES_GROUP_BY_ITEM_RESOLUTION_NODE = "vr"

    DATAFLOW_PLAN_PREFIX = "dfp"
    DATAFLOW_PLAN_SUBGRAPH_PREFIX = "dfpsub"
    OPTIMIZED_DATAFLOW_PLAN_PREFIX = "dfpo"
    SQL_PLAN_PREFIX = "sqp"
    EXEC_PLAN_PREFIX = "ep"

    MF_DAG = "mfd"

    TIME_SPINE_SOURCE = "time_spine_src"
    SUB_QUERY = "subq"
    NODE_RESOLVER_SUB_QUERY = "nr_subq"

    @property
    @override
    def str_value(self) -> str:
        return self.value


@dataclass(frozen=True)
class DynamicIdPrefix(IdPrefix):
    """ID prefixes based on any string value."""

    prefix: str

    @property
    @override
    def str_value(self) -> str:
        return self.prefix



from __future__ import annotations

import collections
from dataclasses import dataclass
from typing import Dict, Optional, Sequence, Tuple

from metricflow_semantics.specs.measure_spec import MeasureSpec
from metricflow_semantics.specs.non_additive_dimension_spec import NonAdditiveDimensionSpec


@dataclass(frozen=True)
class GroupedMeasureSpecsByAdditiveness:
    """Results after grouping measures by their additive properties."""

    grouped_semi_additive_measures: Sequence[Tuple[MeasureSpec, ...]]
    additive_measures: Tuple[MeasureSpec, ...]

    @property
    def measures_by_additiveness(self) -> Dict[Optional[NonAdditiveDimensionSpec], Tuple[MeasureSpec, ...]]:
        """Returns a mapping from additiveness spec to a tuple of measure specs.

        This is useful if you wish to consume the tuples of MeasureSpecs in a single pass without having to
        divide calls up by the existence of an additiveness specification
        """
        additiveness_to_measures: Dict[Optional[NonAdditiveDimensionSpec], Tuple[MeasureSpec, ...]] = {}
        if self.additive_measures:
            additiveness_to_measures[None] = self.additive_measures

        for grouped_specs in self.grouped_semi_additive_measures:
            assert len(grouped_specs) > 0, "received empty set of measure specs, this should not happen!"
            # These all have the same additiveness spec value
            non_additive_spec = grouped_specs[0].non_additive_dimension_spec
            additiveness_to_measures[non_additive_spec] = grouped_specs

        return additiveness_to_measures


def group_measure_specs_by_additiveness(measure_specs: Sequence[MeasureSpec]) -> GroupedMeasureSpecsByAdditiveness:
    """Bucket the provided measure specs by.

    - Additive Measures
    - Semi-additive measures containing the same non-additive dimension attributes
    """
    bucket = collections.defaultdict(list)
    additive_bucket = []
    for spec in measure_specs:
        non_additive_dimension_spec = spec.non_additive_dimension_spec
        if non_additive_dimension_spec:
            bucket[non_additive_dimension_spec.bucket_hash].append(spec)
        else:
            additive_bucket.append(spec)
    return GroupedMeasureSpecsByAdditiveness(
        grouped_semi_additive_measures=tuple(tuple(measures) for measures in bucket.values()),
        additive_measures=tuple(additive_bucket),
    )



from __future__ import annotations


class SqlRenderingConstants:
    """Constants used for rendering SQL."""

    # The indentation for SQL e.g.
    # SELECT
    #   foo
    #   , bar
    INDENT = "  "



"""Helper classes to convert model objects to spec object representations.

In some cases we need to take a model object and convert it to a spec for use in a DataflowPlan or similar.
This fundamentally requires us to combine a model object with a spec object and, in some cases, follow semantic
metadata about linked specs in order to resolve dimension names and the like.

Since model objects and specs should not depend on each other, we do these conversions separately via external
classes, rather than the perhaps more natural approach of adding a to_spec() method on the model objects. These
shims likely point to the need for a bit of an internal refactor, but that's a concern for another time.
"""
from __future__ import annotations

from dbt_semantic_interfaces.protocols.measure import Measure

from metricflow_semantics.specs.measure_spec import MeasureSpec
from metricflow_semantics.specs.non_additive_dimension_spec import NonAdditiveDimensionSpec


class MeasureConverter:
    """Static class for converting Measure model objects to MeasureSpec instances."""

    @staticmethod
    def convert_to_measure_spec(measure: Measure) -> MeasureSpec:
        """Converts a Measure to a MeasureSpec, and properly handles non-additive dimension properties."""
        non_additive_dimension_spec = (
            NonAdditiveDimensionSpec(
                name=measure.non_additive_dimension.name,
                window_choice=measure.non_additive_dimension.window_choice,
                window_groupings=tuple(measure.non_additive_dimension.window_groupings),
            )
            if measure.non_additive_dimension is not None
            else None
        )

        return MeasureSpec(
            element_name=measure.name,
            non_additive_dimension_spec=non_additive_dimension_spec,
        )



from __future__ import annotations

from typing import Optional

from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilter
from dbt_semantic_interfaces.protocols import WhereFilter, WhereFilterIntersection


def merge_to_single_where_filter(where_filter_intersection: WhereFilterIntersection) -> Optional[WhereFilter]:
    """Returns a single where filter that is equivalent to the given intersection."""
    if len(where_filter_intersection.where_filters) == 0:
        return None

    if len(where_filter_intersection.where_filters) == 1:
        return where_filter_intersection.where_filters[0]

    each_where_filter_condition = [
        "( " + where_filter.where_sql_template + " )" for where_filter in where_filter_intersection.where_filters
    ]

    return PydanticWhereFilter(where_sql_template=" AND ".join(each_where_filter_condition))
